"""Class to correct from baseline and chemical noise.
It determines as well the noise and the signal above certain sigma"""

"""___Built-In Modules___"""
from nectar_msi.data_formats.data_spectrum import DataSpectrum
from nectar_msi.noise_correction.noise_determination import NoiseDetermination
from nectar_msi.noise_correction.chemical_noise import ChemicalNoise
from nectar_msi.noise_correction.background_noise import BackgroundNoise
from nectar_msi.plotting.plotting import Plotting

"""___Third-Party Modules___"""
import numpy as np

"""___Authorship___"""
__author__ = "Ariadna Gonzalez-Fernandez"
__email__ = "ariadna.gonzalez@npl.co.uk"


class NoiseCorrection:
    def __init__(self):
        self.noisedetermination = NoiseDetermination()
        self.chemicalnoise = ChemicalNoise()
        self.backgroundnoise = BackgroundNoise()
        self.plotting = Plotting()
        pass

    def noise_correction(self, mean_spectrum, plot_noise=False):
        """
        Function to correct baseline and determine noise level in the mean spectrum

        :param mean_spectrum: mean spectrum to determine the noise
        :type mean_spectrum: DataSpectrum
        :param plot_noise: option to plot the results
        :param plot_noise: Boolean
        :return: mean spectrum baseline corrected and S/N masked
        :rtype: DataSpectrum
        """

        print("Applying baseline correction to the mean spectrum...")
        mean_spectrum_baseline_corr = self.noisedetermination.baseline_correction(mean_spectrum, save_baseline=False)
        intensities_peaks = np.copy(mean_spectrum_baseline_corr.intensities)
        intensities_noise = np.copy(mean_spectrum_baseline_corr.intensities)

        print("Masking the noise...")
        (mask, noiseaveragearray2, noisestvedarray2) = self.noisedetermination.clipandmasklocal_ppmwindow_interpolate(
            mean_spectrum_baseline_corr.intensities, mean_spectrum_baseline_corr.mzs,
            sigma_threshold=3.0, channels=20000)

        intensities_peaks[np.where(mask == 0)] = np.nan  # To choose the peaks of the spectrum according with the S/N
        intensities_noise[np.where(mask == 1)] = np.nan
        mean_spectrum_baseline_corr.set_mzmask(mask)

        if plot_noise:
            self.plotting.plot_noise_spectrum(mean_spectrum_baseline_corr, intensities_peaks, intensities_noise)

        print('Mean spectrum correction done!')
        return mean_spectrum_baseline_corr

    def noise_correction_with_chemical_noise(self, mean_spectrum, plot_noise=False, plot_chemicalnoise=False):
        """
        Function to correct baseline, determine noise level and correct chemical noise in the mean spectrum

        :param mean_spectrum: mean spectrum to determine/correct noise
        :type mean_spectrum: DataSpectrum
        :param plot_noise: Option to plot the signal/noise level in the spectrum
        :type plot_noise: Boolean
        :param plot_chemicalnoise: Option to plot the corrected spectrum and the modelled chemical noise
        :type plot_chemicalnoise: Boolean
        :return: mean spectrum corrected
        :rtype: DataSpectrum
        """

        print("Applying baseline correction...")
        mean_spectrum_baseline_corr = self.noisedetermination.baseline_correction(mean_spectrum, save_baseline=False)

        intensities_peaks = np.copy(mean_spectrum_baseline_corr.intensities)
        intensities_noise = np.copy(mean_spectrum_baseline_corr.intensities)

        print("Masking the noise...")
        (mask, noiseaveragearray, noisestvedarray) = self.noisedetermination.clipandmasklocal_ppmwindow_interpolate(
            mean_spectrum_baseline_corr.intensities, mean_spectrum_baseline_corr.mzs, sigma_threshold=3, channels=20000)

        intensities_peaks[np.where(mask == 0)] = np.nan  # To choose the peaks of the spectrum according with the S/N
        intensities_noise[np.where(mask == 1)] = np.nan
        mean_spectrum_baseline_corr.set_mzmask(mask)

        if plot_noise:
            self.plotting.plot_noise_spectrum(mean_spectrum_baseline_corr, intensities_peaks, intensities_noise)

        # converts noise to a DataSpectrum object
        noise_spectrum = DataSpectrum(-99, [-99, -99], mean_spectrum_baseline_corr.mzs, intensities_noise)

        # chemical noise correction with 45% percentile subtraction
        noise_corrected_spectrum = self.chemicalnoise.chemical_noise_correction_45percentile(
            noise_spectrum, mean_spectrum_baseline_corr, plot_chemicalnoise=plot_chemicalnoise)

        print("Masking the noise after chemical background...")
        (mask2, noiseaveragearray2, noisestvedarray2) = self.noisedetermination.clipandmasklocal_ppmwindow_interpolate(
            noise_corrected_spectrum.intensities, noise_corrected_spectrum.mzs, sigma_threshold=3, channels=20000.0)

        noise_corrected_spectrum.set_mzmask(mask2)

        if plot_noise:
            # To plot the noise and the signal
            noise_corrected_spectrum_peaks = np.copy(noise_corrected_spectrum.intensities)
            noise_corrected_spectrum_noise = np.copy(noise_corrected_spectrum.intensities)
            # To choose the peaks of the spectrum according with the S/N
            noise_corrected_spectrum_peaks[np.where(mask2 == 0)] = np.nan
            noise_corrected_spectrum_noise[np.where(mask2 == 1)] = np.nan

            self.plotting.plot_noise_spectrum(noise_corrected_spectrum, noise_corrected_spectrum_peaks,
                                              noise_corrected_spectrum_noise)

        print('Mean spectrum correction done!')
        return noise_corrected_spectrum



"""nectar_msi - nectar_msi (NoisE CorrecTion AlgoRithm) is a python package for noise correction
in Q-ToF MSI instruments."""

from nectar_msi.data_formats.data_cube import DataCube
from nectar_msi.data_formats.data_spectrum import DataSpectrum
from nectar_msi.data_processing.readers import Readers
from nectar_msi.data_processing.savers import Savers
from nectar_msi.data_processing.data_operations import DataOperations
from nectar_msi.noise_correction.noise_determination import NoiseDetermination
from nectar_msi.noise_correction.noise_correction import NoiseCorrection
from nectar_msi.noise_correction.chemical_noise import ChemicalNoise
from nectar_msi.noise_correction.background_noise import BackgroundNoise
from nectar_msi.database_matching.database_matching import DatabaseMatching
from nectar_msi.plotting.plotting import Plotting
from nectar_msi.peak_picking.peak_picking import PeakPicking

__author__ = "Ariadna Gonzalez-Fernandez <ariadna.gonzalez@npl.co.uk>"
__all__ = []

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions



"""Class to remove the chemical noise following the Adaptive Background Subtraction"""

"""___Built-In Modules___"""
from nectar_msi.plotting.plotting import Plotting
from nectar_msi.data_formats.data_spectrum import DataSpectrum

"""___Third-Party Modules___"""
import numpy as np
import scipy.interpolate

"""___Authorship___"""
__author__ = "Ariadna Gonzalez-Fernandez"
__email__ = "ariadna.gonzalez@npl.co.uk"


class ChemicalNoise:
    def __init__(self):
        self.plotting = Plotting()
        pass

    def chemical_noise_correction_45percentile(self, mean_noise_spectrum, mean_spectrum, plot_chemicalnoise=False):
        """
        Function to apply chemical noise correction in the mean spectrum

        :param mean_noise_spectrum: noise spectrum
        :type mean_noise_spectrum: DataSpectrum
        :param mean_spectrum: baseline corrected mean spectrum
        :type mean_spectrum: DataSpectrum
        :param plot_chemicalnoise: option to plot the original, final and chemical noise spectra
        :type plot_chemicalnoise: Boolean
        :return: corrected mean spectrum
        :type: DataSpectrum

        """
        xaxis_steps = np.arange(mean_spectrum.mzs[0] + 1, mean_spectrum.mzs[-1] - 1, 1)
        intensities_new, chemical_noise = self.adaptive_background_subtraction(
            mean_spectrum, mean_noise_spectrum.intensities, xaxis_steps, percentile=45)

        if plot_chemicalnoise:
            self.plotting.plot_spectra_with_chemicalnoise(mean_spectrum, intensities_new, chemical_noise)

        spectrum_corrected = DataSpectrum(-99, [-99, -99], mean_spectrum.mzs, intensities_new)

        return spectrum_corrected

    def adaptive_background_subtraction(self, mean_spectrum, intensities_noise, xaxis_steps, percentile=45):
        """
        Function to estimate and subtract the chemical noise (Adaptive Background Subtraction approach)

        :param mean_spectrum: baseline corrected mean spectrum
        :type mean_spectrum: DataSpectrum
        :param intensities_noise: intensity values masked as noise
        :type intensities_noise: array of floats
        :param xaxis_steps: size of the x-axis bins
        :type xaxis_steps: array of floats
        :param percentile: percentile to be subtracted (default 45)
        :type percentile: int
        :return: corrected intensities, chemical_noise model intensities
        :type: array, array
        """

        intensities_new = np.zeros_like(mean_spectrum.intensities)  # corrected intensity values
        chemical_noise = np.zeros_like(mean_spectrum.intensities)  # chemical noise model

        for xaxis_current in xaxis_steps:
            bins = np.empty((21, 10))
            xbins = (np.linspace(
                xaxis_current - 10.5, xaxis_current + 10.5, 210, endpoint=False).reshape((21, 10)) + 0.05)
            for kk in range(0, 21, 1):
                for kkk in range(0, 10, 1):
                    bins[kk, kkk] = np.average(intensities_noise[np.where(
                        (mean_spectrum.mzs > xbins[kk, kkk] - 0.05) & (mean_spectrum.mzs <= xbins[kk, kkk] + 0.05))])

            xaxis_centroids = (np.arange(xaxis_current - 0.6, xaxis_current + 0.6, 0.1) + 0.05)
            background_noise = np.nanpercentile(bins, percentile, axis=0)
            background_noise = np.append(background_noise, background_noise[0])
            background_noise = np.insert(background_noise, 0, background_noise[-1])

            interpolfunc = scipy.interpolate.interp1d(
                xaxis_centroids[np.where(np.isfinite(background_noise))],
                background_noise[np.where(np.isfinite(background_noise))],
                bounds_error=False,
                fill_value="extrapolate")

            xaxis_currentbin = mean_spectrum.mzs[np.where(
                (mean_spectrum.mzs > xaxis_current - 0.5) & (mean_spectrum.mzs <= xaxis_current + 0.5))]
            yvals = interpolfunc(xaxis_currentbin)[:]

            chemical_noise[np.where(
                (mean_spectrum.mzs > xaxis_current - 0.5) & (mean_spectrum.mzs <= xaxis_current + 0.5))] = yvals

            intensities_original_1bin = mean_spectrum.intensities[np.where(
                (mean_spectrum.mzs > xaxis_current - 0.5) & (mean_spectrum.mzs <= xaxis_current + 0.5))]

            background_subtraction = intensities_original_1bin - yvals

            intensities_new[np.where(
                (mean_spectrum.mzs > xaxis_current - 0.5) &
                (mean_spectrum.mzs <= xaxis_current + 0.5))] = background_subtraction

        return intensities_new, chemical_noise



"""Class to do the database matching with the HMDB database.
It takes into account the presence of isotopes."""

"""___Built-In Modules___"""

"""___Third-Party Modules___"""
import numpy as np
import pandas as pd

"""___Authorship___"""
__author__ = "Ariadna Gonzalez"
__email__ = "ariadna.gonzalez@npl.co.uk"


class DatabaseMatching:
    def __init__(self):
        pass

    def database_matching(self, list_of_compounds, polarity, modality, path_outputs, ppm=30.0):
        """
        Function to obtain the list of all peaks that match with the HMDB database. The
        information in the output table will be:
        ["hmbdb_id","name","chemical_formula", "monoisotopic_mass","monoisotopic_mass_HMDB", "polarity",
        "mode","adduct", "meas mz","peak intensity","n_lines","n_isotopes",
        "theoretical mass", "ppm", "ppm_abs","kingdom", "super_class","class"]

        :param list_of_compounds: list of peaks
        :type list_of_compounds: pandas DataFrame
        :param polarity: polarity in which the data was acquire
        :type polarity: str
        :param modality: modality in which the data was acquire
        :type modality: str
        :param path_outputs: path to save output table
        :type path_outputs: str
        :param ppm: Absolut value uncertainty on the m/z centroid to cross-match with HMDB
        :type ppm: float
        :return: table with all possible matches
        :rtype: pandas DataFrame
        """

        path_libraries = "C:\\Users\\ag12\\PycharmProjects\\nectar\\HMDB files\\"

        '''Possible adducts for each polarity'''
        if polarity == "positive":
            adducts = {"[M+H]": "1.00727645", "[M+Na]": "22.9892207", "[M+K]": "38.963158", "[M-OH]": "-17.00273965",
                       "[M+H30]": "19.01838971", "[M+NH4]": "18.033823"}  # subtracted electron_mass
        if polarity == "negative":
            adducts = {"[M-H3O]": "-19.01838971", "[M-H]": "-1.007276", "[M+OH]": "17.00273965",
                       "[M+Cl]": "34.969402"}  # added electron_mass electron_mass = 0.00054857990924

        '''initialise final table and read HMDB internal library (this library was updated on 2023)'''
        hmdb_compounds = np.empty((0, 18))
        for keys in adducts.keys():
            library = np.load(path_libraries + "hmdb_isotopes_library_" + polarity + "_" + keys + ".npy",
                              allow_pickle=True)
            print("Finding molecules in the spectrum " + keys + "...")
            for i in range(len(library)):
                hmdb_id = library[i][0]
                name = library[i][1]  # name
                name_form = library[i][2]  # chemical formula
                monoisotopic_mass_hmdb = library[i][3]
                kingdom = library[i][4]
                superclass = library[i][5]
                classs = library[i][6]
                isopeak_location = library[i][7]  # centroid of the isotope
                peak_probability = library[i][8]  # probability of the isotope
                monoisotopic_mass = library[i][9]  # - float(adducts.get(keys))

                AllLinesObserved = True
                threshold = (0.10 * peak_probability[0])  # Probability threshold for the intensities
                for j in range(len(peak_probability)):
                    if peak_probability[j] > threshold:
                        linefound = False

                        for ii in range(len(list_of_compounds["meas mz"])):
                            if (abs(((isopeak_location[j] - float(list_of_compounds.iloc[ii, 1])) /
                                     float(list_of_compounds.iloc[ii, 1])) * 10**6) < ppm):
                                linefound = True
                        if linefound == False:
                            AllLinesObserved = False

                if AllLinesObserved == True:
                    counter = 1
                    for jj in range(len(peak_probability)):
                        if peak_probability[jj] > threshold:
                            linefound = False
                            for ii in range(len(list_of_compounds["meas mz"])):
                                if (abs(((isopeak_location[jj] - float(list_of_compounds.iloc[ii, 1])) /
                                         float(list_of_compounds.iloc[ii, 1]) * 10**6)) < ppm):

                                    ppm_dif = ((isopeak_location[jj] - float(list_of_compounds.iloc[ii, 1])) /
                                               float(list_of_compounds.iloc[ii, 1])) * 10**6
                                    ppm_abs = abs(((isopeak_location[jj] - float(list_of_compounds.iloc[ii, 1])) /
                                                   float(list_of_compounds.iloc[ii, 1]) * 10**6))
                                    linefound = True

                                    hmdb_compounds = np.append(hmdb_compounds, np.array([[
                                                    hmdb_id,
                                                    name,
                                                    name_form,
                                                    monoisotopic_mass,
                                                    monoisotopic_mass_hmdb,
                                                    polarity,
                                                    modality,
                                                    str(keys),
                                                    list_of_compounds.iloc[ii, 1],
                                                    list_of_compounds.iloc[ii, 5],
                                                    counter,
                                                    jj,
                                                    isopeak_location[jj],
                                                    ppm_dif,
                                                    ppm_abs,
                                                    str(kingdom),
                                                    str(superclass),
                                                    str(classs)]]), axis=0)
                                    counter += 1  # Gives me the number of detected isotopes

        hmdb_present_compounds = pd.DataFrame(hmdb_compounds, columns=[
                "hmbdb_id",
                "name",
                "name_form",
                "monoisotopic_mass",
                "monoisotopic_mass_HMDB",
                "polarity",
                "mode",
                "adduct",
                "meas mz",
                "peak intensity",
                "n_lines",
                "n_isotopes",
                "theoretical mass",
                "ppm",
                "ppm_abs",
                "kingdom",
                "super_class",
                "class"])

        hmdb_present_compounds.to_csv(path_outputs + "HMDB_compounds_of_interest.csv", index=False)
        print("! List of peaks saved in: " + path_outputs + "HMDB_compounds_of_interest.csv")

        return




"""Class to determine and mask the noise level on the data, as well as to determine the
scatter in the resolving power"""

"""___Built-In Modules___"""
from nectar_msi.data_formats.data_spectrum import DataSpectrum
from nectar_msi.plotting.plotting import Plotting

"""___Third-Party Modules___"""
import numpy as np
from scipy import interpolate

"""___Authorship___"""
__author__ = "Ariadna Gonzalez-Fernandez"
__email__ = "ariadna.gonzalez@npl.co.uk"


class NoiseDetermination:
    def __init__(self):
        self.plotting = Plotting()
        pass

    def SigmaClip(self, values, tolerance=0.01, median=True, sigma_thresh=3.0, no_zeros=True):
        """
        Function to remove outliers in an array via iteration with a threshold value

        :param values: Values used to apply the function
        :type values: array
        :param tolerance: tolerance value to stop the iteration (default = 0.01)
        :type tolerance: float
        :param median: option to choose the mean or the median (default = median)
        :type median: bool
        :param sigma_thresh: sigma threshold value (default = 3)
        :type sigma_thresh: float
        :param no_zeros: if sigma_new is zero, returns the std of the original array
        :type no_zeros: bool
        :return: [sigma_new, average, values]
        :rtype: float, float, array
        """

        # Remove NaNs from input values
        values = np.array(values)
        values = values[np.where(np.isnan(values) == False)]
        values_original = np.copy(values)
        # Continue loop until result converges
        diff = 10e10
        while diff > tolerance:
            # Assess current input iteration
            if median == False:
                average = np.mean(values)
            elif median == True:
                average = np.median(values)
            sigma_old = np.std(values)

            # Mask those pixels that lie more than 3 stdv away from mean
            check = np.zeros([len(values)])
            check[np.where(values > (average + (sigma_thresh * sigma_old)))] = 1
            values = values[np.where(check < 1)]

            # Re-measure sigma and test for convergence
            sigma_new = np.std(values)
            diff = abs(sigma_old - sigma_new) / sigma_old

        # Perform final mask
        check = np.zeros([len(values)])
        check[np.where(values > (average + (sigma_thresh * sigma_old)))] = 1
        check[np.where(values < (average - (sigma_thresh * sigma_old)))] = 1
        values = values[np.where(check < 1)]

        # If required, check if calculated sigma is zero
        if no_zeros:
            if sigma_new == 0.0:
                sigma_new = np.std(values_original)
                if median == False:
                    average = np.mean(values)
                elif median == True:
                    average = np.median(values)
        return [sigma_new, average, values]

    def sigmaClip_stdv(self, width_corr, mz_corr, tolerance=0.01, sigma_thresh=3):
        """
        Function to define resolving power of the sample. Take all peaks and apply
        SigmaClipping to eliminate outliers from the correlation. (intercept not forced to zero)

        :param width_corr: Values on the y-axis (widths from the gaussian fitting)
        :type width_corr: array
        :param mz_corr: Values on the x-axis (centroids of the peaks)
        :type mz_corr: array
        :param tolerance: tolerance value to stop the iteration (default = 0.01)
        :type tolerance: float
        :param sigma_thresh: sigma threshold value (default = 3)
        :type sigma_thresh: float
        :return: func, distance_std, std_new, mz_corr, width_corr
        :rtype:
        """
        diff = 10e10
        while diff > tolerance:
            a = np.polyfit(mz_corr, width_corr, 1)
            func = np.poly1d(a)
            std_old = np.sqrt(np.sum(((width_corr) - func(mz_corr)) ** 2) / len(width_corr))
            distance_std = (width_corr - func(mz_corr)) / std_old
            mz_corr = mz_corr[np.where(abs(distance_std) < sigma_thresh)]
            width_corr = width_corr[np.where(abs(distance_std) < sigma_thresh)]
            a = np.polyfit(mz_corr, width_corr, 1)
            func = np.poly1d(a)
            std_new = np.sqrt(np.sum(((width_corr) - func(mz_corr)) ** 2) / len(width_corr))
            distance_std = (width_corr - func(mz_corr)) / std_old
            diff = abs(std_old - std_new) / std_old

        return func, distance_std, std_new, mz_corr, width_corr

    def sigmaClip_stdv_zero_intercept(self, width_corr, mz_corr, tolerance=0.01, sigma_thresh=3):
        """
        Function to define resolving power of the sample. Takes most intense 200 peaks and apply
         SigmaClipping to eliminate outliers from the correlation. (intercept of the correlation forced to zero)

        :param width_corr: Values on the y-axis (widths from the gaussian fitting)
        :type width_corr: array
        :param mz_corr: Values on the x-axis (centroids of the peaks)
        :type mz_corr: array
        :param tolerance: tolerance value to stop the iteration (default = 0.01)
        :type tolerance: float
        :param sigma_thresh: sigma threshold value (default = 3)
        :type sigma_thresh: float
        :return: func (correlation), distance_std, std_new, mz_corr (selected values), width_corr (selected values)
        :rtype: poly1d, float, float, array, array
        """

        diff = 10e10
        while diff > tolerance:
            mz_corrT = mz_corr[:, np.newaxis]
            a, _, _, _ = np.linalg.lstsq(mz_corrT, width_corr)
            func = np.poly1d([a[0], 0])
            std_old = np.sqrt(np.sum(((width_corr) - func(mz_corr)) ** 2) / len(width_corr))
            distance_std = (width_corr - func(mz_corr)) / std_old
            mz_corr = mz_corr[np.where(abs(distance_std) < sigma_thresh)]
            width_corr = width_corr[np.where(abs(distance_std) < sigma_thresh)]
            mz_corrT = mz_corr[:, np.newaxis]
            a, _, _, _ = np.linalg.lstsq(mz_corrT, width_corr)
            func = np.poly1d([a[0], 0])
            std_new = np.sqrt(np.sum(((width_corr) - func(mz_corr)) ** 2) / len(width_corr))
            distance_std = (width_corr - func(mz_corr)) / std_old
            diff = abs(std_old - std_new) / std_old

        return func, distance_std, std_new, mz_corr, width_corr

    def clipandmask(self, intensities):
        """
        Function to clip and mask an array with the SigmaClipping function

        :param intensities: values to apply the SigmaClipping function on
        :type intensities: array
        :return: mask for signal and noise, average noise, stv noise
        :rtype: array, float, float
        """

        # Create an empty array to determine the mask noise and signal mask
        flag = np.zeros(len(intensities))
        # Apply SigmaClipping
        noise = self.SigmaClip(intensities)  # outputs [sigma_new, average, values]
        noiseavg = noise[1]
        noisestd = noise[0]
        # Apply the mask
        for k in range(len(intensities)):
            if intensities[k] >= noiseavg + 3.0 * noisestd:
                flag[k] = 1
        return flag, noiseavg, noisestd  # flag is an array with ones for the peaks and zeros for the noise

    def clipandmasklocal_ppmwindow_interpolate(self, intensities, xaxis, sigma_threshold=3.0,
                                               channels=20000.0, interpolation_steps=250):
        """
        Function to apply SigmaClipping within a window that changes size with ppm, and interpolation to make it faster

        :param intensities: values to apply the SigmaClipping function on
        :type intensities: array
        :param xaxis: sliding window
        :type xaxis: array
        :param sigma_threshold: sigma threshold value (default = 3)
        :type sigma_threshold: float
        :param channels: number of channels in the window
        :type channels: float
        :param interpolation_steps: number of steps to do the interpolation
        :type interpolation_steps: int
        :return: mask for signal and noise, average noise, stv noise
        :rtype: array, float, float
        """

        # Sigma clipping function local. Sliding window of 20000bins, with interpolation to make it faster.
        flag = np.zeros(len(intensities))
        noiseavgarrayint = np.zeros(int(len(intensities) / interpolation_steps))
        noisestdarrayint = np.zeros(int(len(intensities) / interpolation_steps))
        xarrayint = np.zeros(int(len(intensities) / interpolation_steps))
        counti = 0

        for k in range(500, len(intensities) - 500, interpolation_steps):
            xarrayint[counti] = xaxis[k]
            noise = self.SigmaClip(intensities[np.where(abs(xaxis - xaxis[k]) < channels * xaxis[k] / 1e6)],
                                   sigma_thresh=sigma_threshold)
            noiseavgarrayint[counti] = noise[1]  # Local noise - average
            noisestdarrayint[counti] = noise[0]  # Local stdv - sigma new
            counti += 1
        favg = interpolate.interp1d(xarrayint, noiseavgarrayint, bounds_error=False)
        fstd = interpolate.interp1d(xarrayint, noisestdarrayint, bounds_error=False)
        noiseavgarray = favg(xaxis)
        noisestdarray = fstd(xaxis)
        flag[np.where(intensities > noiseavgarray + 3.0 * noisestdarray)] = 1

        return flag, noiseavgarray, noisestdarray

    def local_noise_around_peak(self, spectrum, peak, window_size=30):
        """
        Function to calculate the S/N around a peak in a given window size (Dalton)

        :param spectrum: spectrum
        :type spectrum: DataSpectrum
        :param peak: m/z centroid value
        :type peak: float
        :param window_size: size of the window around the peak to determine the local noise
        :type window_size: float
        :return: local noise around the peak in +-30Da
        :rtype: float
        """

        # xaxis_window = spectrum.mzs[np.where((spectrum.mzs > peak - window_size)
        #                                      & (spectrum.mzs < peak + window_size))]

        intensities_window = spectrum.intensities[np.where((spectrum.mzs > peak - window_size)
                                                           & (spectrum.mzs < peak + window_size))]

        mask, noiseaveragearray, local_noise = self.clipandmask(intensities_window)

        return local_noise

    def baseline_correction(self, mean_spectrum, save_baseline=False):
        """
        Function to apply baseline correction to the mean spectrum

        :param mean_spectrum: mean spectrum without baseline correction
        :type mean_spectrum: DataSpectrum
        :param save_baseline: option to save the baseline as a DataSpectrum
        :type save_baseline: Boolean
        :return mean_spectrum_baseline_corr: baseline corrected spectrum
        :rtype mean_spectrum_baseline_corr: DataSpectrum
        """

        # Baseline correction: noiseavgarray is the average noise array, and noisestdarray is the sdtve noise array.
        # Default: Both are calculated by interpolating in a window of 250
        (flag, noiseavgarray, noisestdarray) = self.clipandmasklocal_ppmwindow_interpolate(
            mean_spectrum.intensities, mean_spectrum.mzs, sigma_threshold=2.0, channels=20000)

        mean_spectrum_baseline_corr = DataSpectrum(-99, [-99, -99], mean_spectrum.mzs, mean_spectrum.intensities)
        mean_spectrum_baseline_corr.intensities = (mean_spectrum.intensities - noiseavgarray)

        if save_baseline:
            baseline = DataSpectrum(-99, [-99, -99], mean_spectrum.mzs, noiseavgarray)
            return mean_spectrum_baseline_corr, baseline
        else:
            return mean_spectrum_baseline_corr

    def clipandmasklocal_daltonwindow_interpolate(self, intensities, xaxis, sigma_threshold=3.0, window=30.0,
                                                  interpolation_steps=1):
        """
        Function to apply SigmaClipping within a given window size (Dalton). Interpolation to make it faster

        :param intensities:  values to apply the SigmaClipping function on
        :type intensities: array
        :param xaxis: sliding window
        :type xaxis: array
        :param sigma_threshold: sigma threshold value (default = 3)
        :type sigma_threshold: float
        :param window: window size in Dalton
        :type window: float
        :param interpolation_steps: number of steps to do the interpolation
        :type interpolation_steps: int
        :return: mask for signal and noise, average noise, stv noise
        :rtype: array, float, float
        """
        # Sigma clipping function local. Sliding window of 20000bins, with interpolation to make it faster.
        peaks = range(int(np.ceil(xaxis[0])), int(np.ceil(xaxis[-1])), interpolation_steps)
        flag = np.zeros(len(intensities))
        noiseavgarrayint = np.zeros(len(peaks))
        noisestdarrayint = np.zeros(len(peaks))
        counti = 0
        for peak in peaks:
            intensities_window = intensities[np.where((xaxis > peak - window) & (xaxis < peak + window))]
            noise = self.SigmaClip(intensities_window, sigma_thresh=sigma_threshold)
            noiseavgarrayint[counti] = noise[1]  # Local noise - average
            noisestdarrayint[counti] = noise[0]  # Local stdv - sigma new
            counti += 1
        favg = interpolate.interp1d(peaks, noiseavgarrayint, bounds_error=False)
        fstd = interpolate.interp1d(peaks, noisestdarrayint, bounds_error=False)
        noiseavgarray = favg(xaxis)
        noisestdarray = fstd(xaxis)
        flag[np.where(intensities > noiseavgarray + 3.0 * noisestdarray)] = 1

        return flag, noiseavgarray, noisestdarray



"""Class to define Gaussian fitting functions"""

"""___Built-In Modules___"""

"""___Third-Party Modules___"""
import numpy as np
from scipy.optimize import minimize

"""___Authorship___"""
__author__ = "Ariadna Gonzalez-Fernandez"
__email__ = "ariadna.gonzalez@npl.co.uk"


class GaussianFittingDefinitions:
    def __init__(self):
        pass

    def gaussian(self, centroid, peak_intensity, width, xaxis):
        """
        Function to define a gaussian

        :param centroid: centroid of the Gaussian
        :type centroid: float
        :param peak_intensity: amplitude of the Gaussian
        :type peak_intensity: float
        :param width: width of the Gaussian
        :type width: float
        :param xaxis: x-axis window
        :type xaxis: array
        :return: array of intensities - gaussian curve
        :rtype: array
        """
        gausisintensities = peak_intensity * np.exp(-((xaxis - centroid) ** 2) / (2 * width**2))
        return gausisintensities

    def cal_chi2(self, input_array, intensities, xaxis):
        """
        Function to minimise

        :param input_array: array with centroid, intensity and width of the gaussian
        :type input_array: array
        :param intensities: intensities on the x-axis window
        :type intensities: array of floats
        :param xaxis: x-axis window
        :type xaxis: array of floats
        :return: chi2 to minimise
        :rtype: float
        """

        centroid, peak_intensity, width = input_array
        gaussintensities = self.gaussian(centroid, peak_intensity, width, xaxis)

        return np.sum((gaussintensities - intensities) ** 2)

    def multigaussian(self, centroid_1, peak_intensity_1, width_1, centroid_2, peak_intensity_2, width_2, xaxis):
        """
        Function to defined two overlapping Gaussian

        :param centroid_1: centroid of the first Gaussian
        :type centroid_1: float
        :param peak_intensity_1: amplitude of the first Gaussian
        :type peak_intensity_1: float
        :param width_1: width of the first Gaussian
        :type width_1: float
        :param centroid_2: centroid of the second Gaussian
        :type centroid_2: float
        :param peak_intensity_2: amplitude of the second Gaussian
        :type peak_intensity_2: float
        :param width_2: width of the second Gaussian
        :type width_2: float
        :param xaxis: x-axis window
        :type xaxis: array
        :return: total intensity under the curves of two Gaussian
        :rtype: array
        """

        gausisintensities_1 = peak_intensity_1 * np.exp(-((xaxis - centroid_1) ** 2) / (2 * width_1**2))
        gausisintensities_2 = peak_intensity_2 * np.exp(-((xaxis - centroid_2) ** 2) / (2 * width_2**2))
        multigausisintensities = gausisintensities_1 + gausisintensities_2

        return multigausisintensities

    def cal_chi2_multigaussian(self, input_array, intensities, xaxis):
        """
        Function to minimise

        :param input_array: array with centroid, intensity and width of each gaussian
        :type input_array: array
        :param intensities: intensities on the x-axis window
        :type intensities: array
        :param xaxis: x-axis window
        :type: array
        :return: chi2 to minimise
        :rtype: float
        """

        (centroid_1, peak_intensity_1, width_1, centroid_2, peak_intensity_2, width_2,) = input_array
        multigaussintensities = self.multigaussian(centroid_1, peak_intensity_1, width_1, centroid_2,
                                                   peak_intensity_2, width_2, xaxis)
        return np.sum((multigaussintensities - intensities) ** 2)

    def multigaussian_3g(self, centroid_1, peak_intensity_1, width_1,
                         centroid_2, peak_intensity_2, width_2,
                         centroid_3, peak_intensity_3, width_3, xaxis):
        """
        Function to defined three overlapping Gaussian

        :param centroid_1: centroid of the first Gaussian
        :type centroid_1: float
        :param peak_intensity_1: amplitude of the first Gaussian
        :type peak_intensity_1: float
        :param width_1: width of the first Gaussian
        :type width_1: float
        :param centroid_2: centroid of the second Gaussian
        :type centroid_2: float
        :param peak_intensity_2: amplitude of the second Gaussian
        :type peak_intensity_2: float
        :param width_2: width of the second Gaussian
        :type width_2: float
        :param centroid_3: centroid of the third Gaussian
        :type centroid_3: float
        :param peak_intensity_3: amplitude of the third Gaussian
        :type peak_intensity_3: float
        :param width_3: width of the third Gaussian
        :type width_3: float
        :param xaxis: x-axis window
        :type xaxis: array
        :return: total intensity under the curves of two Gaussian
        :rtype: array
        """

        gausisintensities_1 = peak_intensity_1 * np.exp(-((xaxis - centroid_1) ** 2) / (2 * width_1**2))
        gausisintensities_2 = peak_intensity_2 * np.exp(-((xaxis - centroid_2) ** 2) / (2 * width_2**2))
        gausisintensities_3 = peak_intensity_3 * np.exp(-((xaxis - centroid_3) ** 2) / (2 * width_3**2))

        multigausisintensities_3g = (gausisintensities_1 + gausisintensities_2 + gausisintensities_3)

        return multigausisintensities_3g

    def cal_chi2_multigaussian_3g(self, input_array, intensities, xaxis):
        """
        Function to minimise

        :param input_array: array with centroid, intensity and width of each gaussian
        :type input_array: array
        :param intensities: intensities on the x-axis window
        :type intensities: array
        :param xaxis: x-axis window
        :type: array
        :return: chi2 to minimise
        :rtype: float
        """
        (centroid_1, peak_intensity_1, width_1,
         centroid_2, peak_intensity_2, width_2,
         centroid_3, peak_intensity_3, width_3) = input_array

        multigaussintensities_3g = self.multigaussian_3g(
            centroid_1, peak_intensity_1, width_1,
            centroid_2, peak_intensity_2, width_2,
            centroid_3, peak_intensity_3, width_3, xaxis)

        return np.sum((multigaussintensities_3g - intensities) ** 2)

    def gaussian_fit_one_element(self, xaxis, intensities):
        """
        Function to fit one gaussian without constrains

        :param xaxis: x-axis window where the peak is located
        :type xaxis: array
        :param intensities: intensities inside the x-axis window
        :type intensities: array
        :return: optimised result
        :rtype: class 'scipy.optimize._optimize.OptimizeResult'>
        """

        bnds = ((min(xaxis), max(xaxis)), (0, None), (0, 0.1))
        # cal_chi2: object to minimise; x0: initial guess
        res = minimize(self.cal_chi2, x0=[xaxis[np.argmax(intensities)], np.max(intensities), 0.01],
                       args=(intensities, xaxis), method="TNC", bounds=bnds, options={"maxiter": 50000, "ftol": 0.0001})

        return res

    def gaussian_fit_one_element_separated(self, xaxis, intensities, xaxis_local, intensities_local):
        """
        Function to fit one gaussian without constrains (when two peaks are in the same window)

        :param xaxis: x-axis window where the peak is located
        :type xaxis: array
        :param intensities: intensities inside the x-axis window
        :type intensities: array
        :param xaxis_local: x-axis around the separated peak
        :param intensities_local: intensities around the separated peak
        :return: optimised result
        :rtype: class 'scipy.optimize._optimize.OptimizeResult'>
        """
        bnds = ((min(xaxis), max(xaxis)), (0, None), (0, 0.1))
        res = minimize(self.cal_chi2, x0=[xaxis_local, intensities_local, 0.01],
                       args=(intensities[np.where(abs(xaxis - xaxis_local) < 0.01)],
                             xaxis[np.where(abs(xaxis - xaxis_local) < 0.02)]),
                       method="TNC", bounds=bnds, options={"maxiter": 50000, "ftol": 0.0001})

        return res

    def gaussian_fit_two_elements(self, xaxis, intensities, xaxis_1, intensities_1, xaxis_2, intensities_2):
        """
        Function to fit two gaussian without constrains

        :param xaxis: x-axis window where the peak is located
        :type xaxis: array
        :param intensities: intensities inside the x-axis window
        :type intensities: array
        :param xaxis_1: local x-axis window for the first peak
        :type xaxis_1: array
        :param intensities_1: local intensities window for the first peak
        :type intensities_1: array
        :param xaxis_2: local x-axis window for the second peak
        :type xaxis_2: array
        :param intensities_2: local intensities window for the second peak
        :type intensities_2: array
        :return: optimised result
        :rtype: class 'scipy.optimize._optimize.OptimizeResult'>
        """
        bnds = ((min(xaxis), max(xaxis)), (0, None), (0, 0.1),
                (min(xaxis), max(xaxis)), (0, None), (0, 0.1))
        res = minimize(self.cal_chi2_multigaussian, x0=[xaxis_1, intensities_1, 0.01,
                                                        xaxis_2, intensities_2, 0.01],
                       args=(intensities[np.where(abs(xaxis - xaxis_2) < 0.03)],
                             xaxis[np.where(abs(xaxis - xaxis_2) < 0.03)]),
                       method="TNC", bounds=bnds, options={"maxiter": 50000, "ftol": 0.0001})

        return res

    def gaussian_fit_three_elements(self, xaxis, intensities, xaxis_1, intensities_1, xaxis_2, intensities_2,
                                    xaxis_3, intensities_3):
        """
        Function to fit three gaussian without constrains

        :param xaxis: x-axis window where the peak is located
        :type xaxis: array
        :param intensities: intensities inside the x-axis window
        :type intensities: array
        :param xaxis_1: local x-axis window for the first peak
        :type xaxis_1: array
        :param intensities_1: local intensities window for the first peak
        :type intensities_1: array
        :param xaxis_2: local x-axis window for the second peak
        :type xaxis_2: array
        :param intensities_2: local intensities window for the second peak
        :type intensities_2: array
        :param xaxis_3: local x-axis window for the third peak
        :type xaxis_3: array
        :param intensities_3: local intensities window for the third peak
        :type intensities_3: array
        :return: optimised result
        :rtype: class 'scipy.optimize._optimize.OptimizeResult'>
        """
        bnds = ((min(xaxis), max(xaxis)), (0, None), (0, 0.1),
                (min(xaxis), max(xaxis)), (0, None), (0, 0.1),
                (min(xaxis), max(xaxis)), (0, None), (0, 0.1))

        res = minimize(self.cal_chi2_multigaussian_3g,
                       x0=[xaxis_1, intensities_1, 0.01,
                           xaxis_2, intensities_2, 0.01,
                           xaxis_3, intensities_3, 0.01],
                       args=(intensities, xaxis), method="TNC", bounds=bnds, options={"maxiter": 50000, "ftol": 0.0001})

        return res

    def gaussian_fit_one_element_width_constraint(self, xaxis, intensities, xaxis_max, peak_max, func, std):
        """
        Function to fit one gaussian with constrains on centroid, width and intensity

        :param xaxis: x-axis window where the peak is located
        :type xaxis: array
        :param intensities: intensities inside the x-axis window
        :type intensities: array
        :param xaxis_max: m/z centroid value of the peak
        :type xaxis_max: float
        :param peak_max: intensity value at the m/z centroid xaxis_max
        :type peak_max: float
        :param func: correlation function to determine constrains according to the resolving power
        :type func: poly1d
        :param std: intercept of the resolving power to determine constrains
        :type std: float
        :return: optimised result
        :rtype: class 'scipy.optimize._optimize.OptimizeResult'>
        """
        bnds = ((xaxis_max - func(xaxis_max) - 1 * std, xaxis_max + func(xaxis_max) + 1 * std),
                (peak_max - 0.5 * peak_max, peak_max + 0.5 * peak_max),
                (func(xaxis_max) - 5 * std, func(xaxis_max) + 5 * std))

        peak_fit = intensities[(xaxis > xaxis_max - func(xaxis_max) - 6 * std)
                               & (xaxis < xaxis_max + func(xaxis_max) + 6 * std)]

        xaxis_fit = xaxis[(xaxis > xaxis_max - func(xaxis_max) - 6 * std)
                          & (xaxis < xaxis_max + func(xaxis_max) + 6 * std)]

        res = minimize(self.cal_chi2, x0=[xaxis_max, peak_max, func(xaxis_max)],
                       args=(peak_fit, xaxis_fit), method="TNC", bounds=bnds,
                       options={"maxiter": 50000, "ftol": 0.0001})

        return res

    def gaussian_fit_two_elements_width_constraint(self, xaxis, intensities, xaxis_1, intensities_1,
                                                   xaxis_2, intensities_2, func, std):
        """
        Function to fit two gaussian with constrains on centroid, width and intensity

        :param xaxis: x-axis window where the peak is located
        :type xaxis: array
        :param intensities: intensities inside the x-axis window
        :type intensities: array
        :param xaxis_1: m/z centroid value of the first peak
        :type xaxis_1: float
        :param intensities_1: intensity value at the m/z centroid of the first peak xaxis_1
        :type intensities_1: float
        :param xaxis_2: m/z centroid value of the second peak
        :type xaxis_2: float
        :param intensities_2: intensity value at the m/z centroid of the second peak xaxis_2
        :type intensities_2: float
        :param func: correlation function to determine constrains according to the resolving power
        :type func: poly1d
        :param std: intercept of the resolving power to determine constrains
        :type std: float
        :return: optimised result
        :rtype: class 'scipy.optimize._optimize.OptimizeResult'>
        """

        bnds = ((xaxis_1 - func(xaxis_1) - 1 * std, xaxis_1 + func(xaxis_1) + 1 * std),
                (intensities_1 - 0.5 * intensities_1, intensities_1 + 0.5 * intensities_1),
                (func(xaxis_1) - 5 * std, func(xaxis_1) + 5 * std),
                (xaxis_2 - func(xaxis_2) - 1 * std, xaxis_2 + func(xaxis_2) + 1 * std),
                (intensities_2 - 0.5 * intensities_2, intensities_2 + 0.5 * intensities_2),
                (func(xaxis_2) - 5 * std, func(xaxis_2) + 5 * std))

        peak_fit = intensities[(xaxis > xaxis_1 - func(xaxis_1) - 6 * std)
                               & (xaxis < xaxis_2 + func(xaxis_2) + 6 * std)]

        xaxis_fit = xaxis[(xaxis > xaxis_1 - func(xaxis_1) - 6 * std)
                          & (xaxis < xaxis_2 + func(xaxis_2) + 6 * std)]

        res = minimize(self.cal_chi2_multigaussian, x0=[xaxis_1, intensities_1, func(xaxis_1),
                                                        xaxis_2, intensities_2, func(xaxis_2)],
                       args=(peak_fit, xaxis_fit), method="TNC", bounds=bnds,
                       options={"maxiter": 50000, "ftol": 0.0001})

        return res

    def gaussian_fit_three_elements_width_constraint(self, xaxis, intensities, xaxis_1, intensities_1,
                                                     xaxis_2, intensities_2, xaxis_3, intensities_3, func, std):
        """
        Function to fit three gaussian with constrains on centroid, width and intensity

        :param xaxis: x-axis window where the peak is located
        :type xaxis: array
        :param intensities: intensities inside the x-axis window
        :type intensities: array
        :param xaxis_1: m/z centroid value of the first peak
        :type xaxis_1: float
        :param intensities_1: intensity value at the m/z centroid of the first peak xaxis_1
        :type intensities_1: float
        :param xaxis_2: m/z centroid value of the second peak
        :type xaxis_2: float
        :param intensities_2: intensity value at the m/z centroid of the second peak xaxis_2
        :type intensities_2: float
        :param xaxis_3: m/z centroid value of the third peak
        :type xaxis_3: float
        :param intensities_3: intensity value at the m/z centroid of the third peak xaxis_2
        :type intensities_3: float
        :param func: correlation function to determine constrains according to the resolving power
        :type func: poly1d
        :param std: intercept of the resolving power to determine constrains
        :type std: float
        :return: optimised result
        :rtype: class 'scipy.optimize._optimize.OptimizeResult'>
        """

        bnds = ((xaxis_1 - func(xaxis_1) - 1 * std, xaxis_1 + func(xaxis_1) + 1 * std),
                (intensities_1 - 0.5 * intensities_1, intensities_1 + 0.5 * intensities_1),
                (func(xaxis_1) - 5 * std, func(xaxis_1) + 5 * std),
                (xaxis_2 - func(xaxis_2) - 1 * std, xaxis_2 + func(xaxis_2) + 1 * std),
                (intensities_2 - 0.5 * intensities_2, intensities_2 + 0.5 * intensities_2),
                (func(xaxis_2) - 5 * std, func(xaxis_2) + 5 * std),
                (xaxis_3 - func(xaxis_3) - 1 * std, xaxis_3 + func(xaxis_3) + 1 * std),
                (intensities_3 - 0.5 * intensities_3, intensities_3 + 0.5 * intensities_3),
                (func(xaxis_3) - 5 * std, func(xaxis_3) + 5 * std))

        xaxis_fit = xaxis[(xaxis > xaxis_1 - func(xaxis_1) - 6 * std) & (xaxis < xaxis_3 + func(xaxis_3) + 6 * std)]
        peak_fit = intensities[(xaxis > xaxis_1 - func(xaxis_1) - 6 * std)
                               & (xaxis < xaxis_3 + func(xaxis_3) + 6 * std)]

        res = minimize(self.cal_chi2_multigaussian_3g, x0=[xaxis_1, intensities_1, func(xaxis_1),
                                                           xaxis_2, intensities_2, func(xaxis_2),
                                                           xaxis_3, intensities_3, func(xaxis_3)],
                       args=(peak_fit, xaxis_fit), method="TNC", bounds=bnds,
                       options={"maxiter": 50000, "ftol": 0.0001})

        return res



"""Class for saving data"""

"""___Built-In Modules___"""
from nectar_msi.data_formats.data_spectrum import DataSpectrum
from nectar_msi.data_formats.data_cube import DataCube

"""___Third-Party Modules___"""
import h5py
from pyimzml.ImzMLWriter import ImzMLWriter
from scipy.io import savemat
import numpy as np
from tqdm import tqdm
import copy


"""___Authorship___"""
__author__ = "Ariadna Gonzalez-Fernandez"
__email__ = "ariadna.gonzalez@npl.co.uk"


class Savers:
    def __init__(self):
        pass

    def save_imzml(self, path_outputs, data_cube):
        """
        Function to save data in imzML format

        :param path_outputs: path to save the data_cube
        :type path_outputs: string
        :param data_cube: data_cube to be saved
        :type data_cube: DataCube
        :return: None
        :rtype: None
        """
        print("Saving imzML file...")
        with ImzMLWriter(path_outputs) as writer:
            for index in range(len(data_cube.coord)):
                spec = data_cube.get_spectrum_idx(data_cube.idxs[index])
                if spec is not None:
                    writer.addSpectrum(spec.mzs, spec.intensities,
                                       (spec.coordinates[0] + 1, spec.coordinates[1] + 1, 1))
        print("imzML saved in " + path_outputs)

    def save_hdf5(self, path_outputs, data_cube):
        """
        Function to save a data_cube into hdf5 format

        :param path_outputs: path to save the data_cube
        :type path_outputs: string
        :param data_cube: data_cube to be saved
        :type data_cube: DataCube
        :return: None
        :rtype: None
        """

        print("Saving hdf5 file...")
        f = h5py.File(path_outputs, "w")
        f.create_dataset("pixelmask", data=data_cube.pixelmask)
        f.create_dataset("idxs", data=data_cube.idxs)
        grp_meta = f.create_group("metadata")
        for key, value in data_cube.cubemetadata.items():
            grp_meta.attrs[key] = value
        grp_data = f.create_group("data")
        grp_data.attrs["shape"] = data_cube.shape
        for index in range(len(data_cube.coord)):
            spec = data_cube.get_spectrum_idx(data_cube.idxs[index])
            if spec is not None:
                grp_spec = grp_data.create_group("spectrum_%s" % data_cube.idxs[index])
                grp_spec.create_dataset("mzs", data=spec.mzs)
                grp_spec.create_dataset("intensities", data=spec.intensities)
                grp_spec.create_dataset("mzmask", data=spec.mzmask)
                grp_spec.attrs["coordinates"] = spec.coordinates
                for key, value in spec.spectrummetadata.items():
                    grp_spec.attrs[key] = value
                for key, value in spec.offsets.items():
                    grp_spec.attrs[key] = value
                del spec
        print("HDF5 saved in " + path_outputs)

    def save_spectrum_hdf5(self, path_outputs, spectrum):
        """
        Function to save a spectrum into hdf5 format

        :param path_outputs: path to save the spectrum
        :type path_outputs: string
        :param spectrum: spectrum to be saved
        :type spectrum: DataSpectrum
        :return: None
        :rtype: None
        """

        print("Saving hdf5 file...")
        f = h5py.File(path_outputs, "w")
        grp_spec = f.create_group("spectrum_data")
        grp_spec.create_dataset("mzs", data=spectrum.mzs)
        grp_spec.create_dataset("intensities", data=spectrum.intensities)
        grp_spec.create_dataset("mzmask", data=spectrum.mzmask)
        grp_spec.attrs["coordinates"] = spectrum.coordinates
        grp_spec.attrs["idx"] = spectrum.idx
        for key, value in spectrum.spectrummetadata.items():
            grp_spec.attrs[key] = value
        print("HDF5 saved in " + path_outputs)

    def save_final_DataCube(self, data_cube, selected_peaks, path_outputs,
                            save_imzml=True, save_hdf5=False, save_mat=False):
        """
        Function to save the de-noised datacube

        :param data_cube: data cube to be saved
        :type data_cube: DataCube
        :param selected_peaks: peaks to be included in the final datacube
        :type selected_peaks: pandas dataframe
        :param path_outputs: path where to save the datacube
        :type path_outputs: str
        :param save_imzml: option to save imzML format
        :type save_imzml: bool
        :param save_hdf5: option to save hdf5 format
        :type save_hdf5: bool
        :param save_mat: option to save .mat format
        :type save_mat: bool
        :return: None
        :rtype: None
        """

        print('Saving final datacube...')
        data_cube_final = DataCube(data_cube.shape, metadata=data_cube.cubemetadata)
        data_cube_final.coord = copy.deepcopy(data_cube.coord)
        data_cube_final.pixelmask = copy.deepcopy(data_cube.pixelmask)
        data_cube_final.idxs = copy.deepcopy(data_cube.idxs)

        data_cube_matfile = np.zeros([len(data_cube.idxs), len(selected_peaks['left min'])])
        peaks = selected_peaks['meas mz'].to_numpy()

        for index in tqdm(range(len(data_cube.idxs))):
            idx = data_cube.idxs[index]
            spectrum_per_pixel = data_cube.get_spectrum_idx(idx)

            xaxis = spectrum_per_pixel.mzs[:]
            intensities = spectrum_per_pixel.intensities[:]
            new_intensities = np.zeros(len(intensities))
            add_intensities = np.zeros(len(selected_peaks['left min']))

            for mzi in range(len(selected_peaks['left min'])):
                new_intensities[
                    np.where((xaxis >= selected_peaks.iloc[mzi, 0]) & (xaxis <= selected_peaks.iloc[mzi, 2]))] = \
                    intensities[
                        np.where((xaxis >= selected_peaks.iloc[mzi, 0]) & (xaxis <= selected_peaks.iloc[mzi, 2]))]

                add_intensities[mzi] = spectrum_per_pixel.get_intensity_between_mzs(selected_peaks.iloc[mzi, 0],
                                                                                    selected_peaks.iloc[mzi, 2])
            data_cube_matfile[index, :] = add_intensities
            spectrum_per_pixel.set_intensities(new_intensities)
            data_cube_final.set_spectrum_idx(idx, spectrum_per_pixel)

        if save_imzml:
            self.save_imzml(path_outputs + 'datacube_reduced.imzML', data_cube_final)
            print('datacube_reduced.imzML saved in ' + path_outputs)
        if save_hdf5:
            self.save_hdf5(path_outputs + 'datacube_reduced.hdf5', data_cube_final)
            print('datacube_reduced.hdf5 saved in ' + path_outputs)
        if save_mat:
            mdic = {"data_cube": data_cube_matfile, "selected_peaks": peaks}
            savemat(path_outputs + "datacube_reduced.mat", mdic)
            print('datacube_reduced.mat saved in ' + path_outputs)




"""Class for reading data"""

"""___Built-In Modules___"""
from nectar_msi.data_formats.data_cube import DataCube
from nectar_msi.data_formats.data_spectrum import DataSpectrum

"""___Third-Party Modules___"""
from pyimzml.ImzMLParser import ImzMLParser
import h5py

"""___Authorship___"""
__author__ = "Ariadna Gonzalez-Fernandez"
__email__ = "ariadna.gonzalez@npl.co.uk"

class Readers:
    def __init__(self):
        pass

    def read_imzml(self, filepath):
        """
        Function to read imzML data and convert it into DataCube class

        :param filepath: path where the data are located
        :type filepath: string
        :return: data_cube object
        :rtype: DataCube
        """
        p = ImzMLParser(filepath)
        dict = p.imzmldict
        x_shape = dict.get("max count of pixels x")
        y_shape = dict.get("max count of pixels y")
        shape = (x_shape, y_shape)
        datac = DataCube(shape)
        print("Image shape", x_shape, y_shape)
        print("Reading imzML file at: " + filepath)
        for idx, (x, y, z) in enumerate(p.coordinates):
            mzs, intensities = p.getspectrum(idx)
            spectrum = DataSpectrum(idx, [x - 1, y - 1], mzs, intensities)
            datac.set_spectrum_idx(idx, spectrum)
            del spectrum
        print("Reading imzML done!")
        return datac

    def read_imzml_missing(self, filepath):
        """
        Function to read imzML data with missing data points (NaNs)

        :param filepath: path where the data are located
        :type filepath: string
        :return: data_cube object
        :rtype: DataCube
        """
        p = ImzMLParser(filepath)
        dict = p.imzmldict
        x_shape = dict.get("max count of pixels x")
        y_shape = dict.get("max count of pixels y")
        shape = (x_shape, y_shape)
        datac = DataCube(shape)
        print("Image shape", x_shape, y_shape)
        print("Reading imzML file at: " + filepath)
        for indexp, (x, y, z) in enumerate(p.coordinates):
            mzs, intensities = p.getspectrum(indexp)
            idx = datac.get_idx_from_coor([x - 1, y - 1])
            spectrum = DataSpectrum(idx, [x - 1, y - 1], mzs, intensities)
            datac.set_spectrum_idx(idx, spectrum)
            del spectrum
        print("Reading imzML done!")
        return datac

    def read_hdf5(self, filepath):
        """
        Function to read hdf5 datacubes

        :param filepath: path where the data is located
        :type filepath: string
        :return: data_cube object
        :rtype: DataCube object
        """
        print("Reading hdf5 file at: " + filepath)
        f = h5py.File(filepath, "r")
        datac = DataCube(f["data"].attrs["shape"], f["idxs"], f["pixelmask"], f["metadata"].attrs)
        for idx in datac.idxs:
            speckey = "spectrum_%s" % idx
            metadata = {}
            offsets = {}
            for key, value in f["data"][speckey].attrs.items():
                if key == "coordinates":
                    coordinates = value
                elif "offset" in key:
                    offsets[key] = value
                else:
                    metadata[key] = value
            spectrum = DataSpectrum(idx, coordinates, f["data"][speckey]["mzs"], f["data"][speckey]["intensities"],
                                    f["data"][speckey]["mzmask"], offsets, metadata)
            datac.set_spectrum_idx(idx, spectrum)
            del spectrum
        print("Reading hdf5 done!")
        return datac

    def read_hdf5_spectrum(self, filepath):
        """
        Function to read a hdf5 spectrum

        :param filepath: path where the spectrum is located
        :type filepath: string
        :return: spectrum object
        :rtype: DataSpectrum
        """
        print("Reading hdf5 file at: " + filepath)
        f = h5py.File(filepath, "r")
        metadata = {}
        offsets = {}
        for key, value in f["spectrum_data"].attrs.items():
            if key == "idx":
                idx = value
            elif key == "coordinates":
                coordinates = value
            elif "offset" in key:
                offsets[key[7::]] = value
            else:
                metadata[key] = value
        spectrum = DataSpectrum(idx, coordinates, f["spectrum_data"]["mzs"],
                                f["spectrum_data"]["intensities"], f["spectrum_data"]["mzmask"], offsets, metadata)
        print("Reading hdf5 done!")
        return spectrum



"""Class for creating and reading data_cubes"""

"""___Third-Party Modules___"""
import numpy as np

"""___Authorship___"""
__author__ = "Ariadna Gonzalez-Fernandez"
__email__ = "ariadna.gonzalez@npl.co.uk"


class DataCube:
    def __init__(self, shape, idxs=None, pixelmask=None, metadata={}):
        """
        Datacube class initialiser

        :param shape: shape of the datacube
        :type shape: tuple
        :param idxs: number of ids of the spectra in the datacube (defaults to None,
        in which case uses the shape to determine number)
        :type idxs: int (optional)
        :param pixelmask: mask to indicate Regions Of Interest (defaults to zeros)
        :type pixelmask: array of int (optional)
        :param metadata: NOT DEVELOPED - Metadata of the cube (default to empty)
        :type metadata: dict (optional)
        """
        self.shape = shape
        self.data = np.empty((shape[0] * shape[1]), dtype=object)
        self.coord = np.array([(i, j) for j in range(shape[1]) for i in range(shape[0])])
        if idxs is None:
            self.idxs = np.arange(shape[0] * shape[1])
        else:
            self.idxs = np.array(idxs, dtype="int")
        if pixelmask is None:
            self.pixelmask = np.zeros(shape[0] * shape[1])
        else:
            self.pixelmask = np.array(pixelmask, dtype="u2")
        self.cubemetadata = metadata

    def get_coor_from_idx(self, idx):
        """
        Function to obtain the coordinates of the spectrum of interest

        :param idx: id of the spectra in the data_cube
        :type idx: int
        :return: coordinates in the data_cube
        :rtype: tuple
        """
        index = np.argmin(abs(idx - self.idxs))
        return self.coord[index]

    def get_idx_from_coor(self, coor):
        """
        Function to obtain the idx given the coordinates of the pixel in the data_cube

        :param coor: coordinates in the data_cube
        :type coor: np.array
        :return: id of the spectrum
        :rtype: int
        """
        return int(self.idxs[np.where((self.coord[:, 0] == coor[0]) & (self.coord[:, 1] == coor[1]))])

    def get_index_from_coor(self, coor):
        """
        Function to obtain the index given the coordinates of the pixel in the data_cube

        :param coor: coordinates in the data_cube
        :type coor: np.array
        :return: index of the array
        :rtype: int
        """
        return coor[0] + coor[1] * self.shape[0]

    def get_spectrum_coor(self, coor):
        """
        Function to obtain the spectrum of a pixel given the coordinates

        :param coor: coordinates of the pixel in the data_cube
        :type coor: tuple
        :return: spectrum of certain pixel
        :rtype: object
        """
        index = self.get_index_from_coor(coor)
        return self.data[index]

    def get_spectrum_idx(self, idx):
        """
        Function to obtain the spectrum of a given idx in the data_cube

        :param idx: idx of the data_cube
        :type idx: int
        :return: spectrum of certain idx
        :rtype: DataSpectrum object
        """
        try:
            return self.data[idx == self.idxs][0]
        except:
            print("The idx given does not seem to be valid.")

    def set_spectrum_coor(self, coor, spectrum):
        """
        Function to set a new spectrum in an specific location given in coordinates

        :param coor: coordinates where to set the new spectrum
        :type coor: tuple
        :param spectrum: spectrum to set in the data_cube
        :type spectrum: DataSpectrum object
        :return: modified data_cube
        :rtype: DataCube
        """
        index = self.get_index_from_coor(coor)
        self.data[index] = spectrum

    def set_spectrum_idx(self, idx, spectrum):
        """
        Function to set a new spectrum in an specific idx location

        :param idx: idx where to set the new spectrum
        :type idx: int
        :param spectrum: spectrum to set in the data_cube
        :type spectrum: DataSpectrum object
        :return: modified data_cube
        :rtype: DataCube
        """
        index = np.argmin(abs(idx - self.idxs))
        self.data[index] = spectrum

    def set_mzs_coor(self, idx, mzs_new):
        """
        Function to set new mz values in a specific pixel selected by its coordinates

        :param idx: id where to set the new mz values
        :type idx: int
        :param mzs_new: new mz values
        :type mzs_new: array
        :return: modified data_cube
        :rtype: DataCube
        """
        index = np.argmin(abs(idx - self.idxs))
        self.data[index].set_mzs(mzs_new)

    def set_mzs_idx(self, idx, mzs_new):
        """
        Function to set new mz values in a specific pixel selected by its idx

        :param idx: idx of the target pixel in the data_cube
        :type idx: int
        :param mzs_new: new mz values
        :type mzs_new: array
        :return: modified data_cube
        :rtype: DataCube
        """
        index = np.argmin(abs(idx - self.idxs))
        self.data[index].set_mzs(mzs_new)

    def set_pixel_mask_1D(self, new_mask):
        """
        Function to set the mask in the data_cube

        :param new_mask: mask to be set
        :type new_mask: intarray
        :return: data_cube with mask
        :rtype: DataCube
        """
        if new_mask.shape == self.pixelmask.shape:
            self.pixelmask = np.array(new_mask, dtype="u2")
        else:
            print("Error: Mask has the wrong shape")

    def set_pixel_mask_2D(self, new_mask):
        """
        Function to set the mask in the data_cube

        :param new_mask: mask to be set
        :type new_mask: intarray
        :return: data_cube with mask
        :rtype: DataCube
        """
        if new_mask.shape[0] == self.shape[0] and new_mask.shape[1] == self.shape[1]:
            self.pixelmask = new_mask.T.flatten()
        else:
            print("Error: Mask has the wrong shape")

    def set_pixel_mask_idx(self, idx, new_mask):
        """
        Function to set the mask in a specific pixel

        :param idx:  idx of the target pixel in the data_cube
        :type idx: int
        :param new_mask: mask to be set
        :type new_mask: intarray
        :return: data_cube with mask
        :rtype: DataCube
        """
        index = np.argmin(abs(idx - self.idxs))
        self.pixelmask[index] = new_mask

    def add_metadata(self, key, value):
        """
        Function to include the metadata in the data_cube (NOT TESTED)

        :param key: attribute of the dictionary
        :type key: string
        :param value: value of the attribute
        :type value: string
        :return: dictionary of the metadata
        :rtype: dict
        """
        self.cubemetadata[key] = value

    def get_pixel_mask_1D(self):
        """
        Function to obtain the mask in 1D

        :return: mask
        :rtype: intarray
        """
        return self.pixelmask

    def get_pixel_mask_2D(self):
        """
        Function to obtain the mask in 2D (image shape)

        :return: mask
        :rtype: intarray
        """
        mask = np.zeros(self.shape)
        for i in range(len(self.coord)):
            mask[self.coord[i][0], self.coord[i][1]] = self.pixelmask[i]
        return mask

    def get_intensities_at_mz(self, mz):
        """
        Function to read the intensity in each spectra for a given mz value to create
        the single ion images, for instance. The intensity is a single reading in each spectra

        :param mz: mz value of interest
        :type mz: float
        :return: Array with the intensity values at the mz location
        :rtype: 2D array
        """
        intensities = np.zeros(self.shape)
        for i in range(len(self.coord)):
            intensities[self.coord[i][0], self.coord[i][1]] = self.data[i].get_intensity_at_mz(mz)
        return intensities

    def get_intensities_between_mzs(self, mzlow, mzhigh):
        """
        Function to obtain the intensities between two values in each spectra to create
        single ion images, for instance. The intensity is the sum between these two values.

        :param mzlow: left value of the interval
        :type mzlow: float
        :param mzhigh: right value of the interval
        :type mzhigh: float
        :return: Array with the intensity values between the interval limits
        :rtype: 2D array
        """
        intensities = np.zeros(self.shape)
        for i in range(len(self.coord)):
            if self.data[i] is not None:
                intensities[self.coord[i][0], self.coord[i][1]] = self.data[i].get_intensity_between_mzs(mzlow, mzhigh)
            else:
                intensities[self.coord[i][0], self.coord[i][1]] = np.nan
        return intensities

    def get_offsets_at_mz(self, mz):
        """
        Function to obtain the offset in each pixel of the data_cube for a given mz reference value

        :param mz: centroid of the peak of interest
        :type mz: array or list
        :return: array with the offset values
        :rtype: 2D array
        """
        offsets = np.zeros(self.shape)
        for i in range(len(self.coord)):
            if self.data[1] is not None:
                offsets[self.coord[i][0], self.coord[i][1]] = self.data[i].get_offset(mz)
            else:
                offsets[self.coord[i][0], self.coord[i][1]] = np.nan
        return offsets



"""Class for Gaussian fitting of peaks"""

"""___Built-In Modules___"""
from nectar_msi.plotting.plotting import Plotting
from nectar_msi.gaussian_fitting.gaussian_fitting_definitions import GaussianFittingDefinitions
from nectar_msi.data_processing.data_operations import DataOperations
from nectar_msi.noise_correction.noise_determination import NoiseDetermination

"""___Third-Party Modules___"""
import numpy as np
import os

"""___Authorship___"""
__author__ = "Ariadna Gonzalez-Fernandez"
__email__ = "ariadna.gonzalez@npl.co.uk"


class GaussianFitting:
    def __init__(self):
        self.plotting = Plotting()
        self.gaussian_def = GaussianFittingDefinitions()
        self.dataop = DataOperations()
        self.noisedetermination = NoiseDetermination()
        pass

    def gaussian_fitting_spectrum(self, mean_spectrum, list_of_peaks, path_outputs, save_fitting=False):
        """
        Function to fit a Gaussian by chi2 minimization. The function saves plots of Gaussian fitting.

        :param mean_spectrum: mean spectrum to take as reference to fit the Gaussian
        :type mean_spectrum: DataSpectrum
        :param list_of_peaks: list of peaks of interest (list created in peak_picking.py)
        :type list_of_peaks: list
        :param path_outputs: outputs path
        :type path_outputs: string
        :param save_fitting: option to save plots of the Gaussian fitting (default = False)
        :type save_fitting: bool
        :return: table with the list of peaks and its characteristics
        (centroid, intensity, width, chi2/Nsamples, Number of sample bins, type of peak,
        success in the fitting)
        :rtype: array
        """

        if save_fitting:
            '''It creates a file where it is going to save all the relevant values of the fitting'''
            dirName = path_outputs + "Gaussian_fitting_most_intense_peaks"
            if not os.path.exists(dirName):
                os.mkdir(dirName)

        # It selects the part of the mean spectrum that has been identified as signal previously
        mask = mean_spectrum.mzmask
        list_of_fittings = np.empty((0, 7))
        # centroid   peak_intensity  width   chi2/Nsamples   Nsamples    Peak_type   Fitting_success
        # 0         1               2       3               4           5           6
        for i in range(2, len(mean_spectrum.intensities) - 2):
            # It selects windows of signal to determine how many peaks are on them
            if mask[i] == 0.0 and mask[i + 1] == 1.0:
                istart = i
            if mask[i] == 1.0 and mask[i + 1] == 0.0:
                spectrum_window2 = mean_spectrum.intensities[istart: i + 2]
                xaxis_window2 = mean_spectrum.mzs[istart: i + 2]

                local_max = []
                # It finds the number of local max in the window to determine how many
                # peaks are in the window.
                for ii in range(len(list_of_peaks)):
                    if max(xaxis_window2) > list_of_peaks[ii] > min(xaxis_window2):
                        local_max = np.append(local_max, list_of_peaks[ii])

                '''Fits one peak'''
                if len(local_max) == 1:
                    res = self.gaussian_def.gaussian_fit_one_element(xaxis_window2, spectrum_window2)
                    if save_fitting:
                        self.plotting.plot_one_gaussian(xaxis_window2, spectrum_window2, res, path_outputs)

                    list_of_fittings = np.append(list_of_fittings,
                                                 np.array([[float(res.x[0]), res.x[1],
                                                            abs(res.x[2]), (self.gaussian_def.cal_chi2
                                                                            (res.x, spectrum_window2, xaxis_window2) /
                                                                            res.x[1]/len(spectrum_window2)),
                                                            len(spectrum_window2), "single", res.success]]), axis=0)

                # If there are two peaks in the window, it determines if they are too
                # close to each other. It they are 15ppm apart, it fits each peak
                # as independent peaks. If not it fits them together as there might be
                # contamination between them.

                if len(local_max) == 2:
                    if abs(local_max[1] - local_max[0]) >= 15:
                        '''Fits two separated peaks'''
                        for j in range(0, 2, 1):
                            peak_max = spectrum_window2[np.where(local_max[j] == xaxis_window2)]
                            xaxis2_max = local_max[j]

                            res = self.gaussian_def.gaussian_fit_one_elements_separated(
                                xaxis_window2, spectrum_window2, xaxis2_max, peak_max)
                            if save_fitting:
                                self.plotting.plot_one_gaussian_separated(xaxis_window2, spectrum_window2,
                                                                         xaxis2_max, peak_max, res, path_outputs)
                            '''Add parameters to final table'''
                            list_of_fittings = np.append(list_of_fittings,
                                                         np.array([[float(res.x[0]), res.x[1], abs(res.x[2]),
                                                                    self.gaussian_def.cal_chi2(
                                                                        res.x, spectrum_window2, xaxis_window2)
                                                                    / res.x[1] / len(spectrum_window2),
                                                                    len(spectrum_window2), "single", res.success,]]),
                                                         axis=0)
                    else:
                        '''Fits two peaks together'''
                        res = self.gaussian_def.gaussian_fit_two_elements(
                            xaxis_window2, spectrum_window2,
                            local_max[0], spectrum_window2[np.where(local_max[0] == xaxis_window2)[0][0]],
                            local_max[1], spectrum_window2[np.where(local_max[1] == xaxis_window2)[0][0]])
                        if save_fitting:
                            self.plotting.plot_two_gaussians(xaxis_window2, spectrum_window2, local_max[0],
                                                             spectrum_window2[np.where(
                                                                 local_max[0] == xaxis_window2)[0][0]], local_max[1],
                                                             spectrum_window2[np.where(
                                                                 local_max[1] == xaxis_window2)[0][0]],
                                                             res, path_outputs)
                        '''Add parameters to final table'''
                        list_of_fittings = np.append(
                            list_of_fittings, np.array([[
                                        float(res.x[0]), res.x[1], abs(res.x[2]),
                                        self.gaussian_def.cal_chi2_multigaussian(res.x, spectrum_window2, xaxis_window2)
                                        / res.x[1] / len(spectrum_window2), len(spectrum_window2), "multiple",
                                res.success]]), axis=0)

                        list_of_fittings = np.append(
                            list_of_fittings, np.array([[
                                        float(res.x[3]), res.x[4], abs(res.x[5]),
                                        self.gaussian_def.cal_chi2_multigaussian(res.x, spectrum_window2, xaxis_window2)
                                        / res.x[4] / len(spectrum_window2), len(spectrum_window2), "multiple",
                                res.success]]), axis=0)

                # If there are tree peaks in the window, it determines the distance among
                # them and fit them accordingly.
                if len(local_max) == 3:
                    if len(local_max) == 3:
                        '''Fits three peaks together'''
                        if np.abs(local_max[1] - local_max[0]) <= 15 and np.abs(local_max[2] - local_max[1]) <= 15:
                            res = self.gaussian_def.gaussian_fit_three_elements(
                                xaxis_window2, spectrum_window2, local_max[0],
                                spectrum_window2[np.where(local_max[0] == xaxis_window2)[0][0]], local_max[1],
                                spectrum_window2[np.where(local_max[1] == xaxis_window2)[0][0]], local_max[2],
                                spectrum_window2[np.where(local_max[2] == xaxis_window2)[0][0]])

                            if save_fitting:
                                self.plotting.plot_three_gausssians(
                                    xaxis_window2, spectrum_window2, local_max[0],
                                    spectrum_window2[np.where(local_max[0] == xaxis_window2)[0][0]],
                                    local_max[1], spectrum_window2[np.where(local_max[1] == xaxis_window2)[0][0]],
                                    local_max[2], spectrum_window2[np.where(local_max[2] == xaxis_window2)[0][0]],
                                    res, path_outputs)
                            '''Add parameters to final table'''
                            list_of_fittings = np.append(
                                list_of_fittings, np.array([[
                                            float(res.x[0]), res.x[1], abs(res.x[2]),
                                            self.gaussian_def.cal_chi2_multigaussian_3g(
                                                res.x, spectrum_window2, xaxis_window2)
                                            / res.x[1] / len(spectrum_window2), len(spectrum_window2), "multiple",
                                            res.success]]), axis=0)

                            list_of_fittings = np.append(
                                list_of_fittings, np.array([[
                                            float(res.x[3]), res.x[4], abs(res.x[5]),
                                            self.gaussian_def.cal_chi2_multigaussian_3g(
                                                res.x, spectrum_window2, xaxis_window2)
                                            / res.x[4] / len(spectrum_window2), len(spectrum_window2), "multiple",
                                            res.success]]), axis=0)

                            list_of_fittings = np.append(
                                list_of_fittings, np.array([[
                                            float(res.x[6]), res.x[7], abs(res.x[8]),
                                            self.gaussian_def.cal_chi2_multigaussian_3g(
                                                res.x, spectrum_window2, xaxis_window2)
                                            / res.x[7] / len(spectrum_window2), len(spectrum_window2), "multiple",
                                            res.success]]), axis=0)

                        '''Fits two peaks together and one separated'''
                        if np.abs(local_max[1] - local_max[0]) <= 15 and np.abs(local_max[2] - local_max[1]) > 15:
                            res = self.gaussian_def.gaussian_fit_two_elements(
                                xaxis_window2, spectrum_window2, local_max[0],
                                spectrum_window2[np.where(local_max[0] == xaxis_window2)[0][0]], local_max[1],
                                spectrum_window2[np.where(local_max[1] == xaxis_window2)[0][0]])
                            if save_fitting:
                                self.plotting.plot_two_gaussians(
                                    xaxis_window2, spectrum_window2, local_max[0],
                                    spectrum_window2[np.where(local_max[0] == xaxis_window2)[0][0]],
                                    local_max[1], spectrum_window2[np.where(local_max[1] == xaxis_window2)[0][0]],
                                    res, path_outputs)

                            '''Add parameters to final table'''
                            list_of_fittings = np.append(
                                list_of_fittings, np.array([[
                                    float(res.x[0]), res.x[1], abs(res.x[2]),
                                    self.gaussian_def.cal_chi2_multigaussian(res.x, spectrum_window2, xaxis_window2)
                                    / res.x[1] / len(spectrum_window2), len(spectrum_window2), "multiple",
                                    res.success]]), axis=0)

                            list_of_fittings = np.append(
                                list_of_fittings, np.array([[
                                    float(res.x[3]), res.x[4], abs(res.x[5]),
                                    self.gaussian_def.cal_chi2_multigaussian(res.x, spectrum_window2, xaxis_window2)
                                    / res.x[4] / len(spectrum_window2), len(spectrum_window2), "multiple",
                                    res.success]]), axis=0)
                            '''Separated peak'''
                            res = self.gaussian_def.gaussian_fit_one_element_separated(
                                xaxis_window2, spectrum_window2, local_max[2],
                                spectrum_window2[np.where(local_max[2] == xaxis_window2)[0][0]])
                            if save_fitting:
                                self.plotting.plot_one_gaussian_separated(
                                    xaxis_window2, spectrum_window2, local_max[2], local_max[2], res, path_outputs)

                            '''Add parameters to final table'''
                            list_of_fittings = np.append(
                                list_of_fittings, np.array([[
                                    float(res.x[0]), res.x[1], abs(res.x[2]),
                                    self.gaussian_def.cal_chi2(res.x, spectrum_window2, xaxis_window2)
                                    / res.x[1] / len(spectrum_window2), len(spectrum_window2), "separated",
                                    res.success]]), axis=0)

                        # The last two peaks are together, they are fitted together. the first one is separated.
                        if np.abs(local_max[2] - local_max[1]) <= 15 and np.abs(local_max[1] - local_max[0]) > 15:
                            res = self.gaussian_def.gaussian_fit_two_elements(
                                xaxis_window2, spectrum_window2,
                                local_max[1], spectrum_window2[np.where(local_max[1] == xaxis_window2)[0][0]],
                                local_max[2], spectrum_window2[np.where(local_max[2] == xaxis_window2)[0][0]])

                            if save_fitting:
                                self.plotting.plot_two_gaussians(
                                    xaxis_window2, spectrum_window2,
                                    local_max[1], spectrum_window2[np.where(local_max[1] == xaxis_window2)[0][0]],
                                    local_max[2], spectrum_window2[np.where(local_max[2] == xaxis_window2)],
                                    res, path_outputs)

                            '''Add parameters to final table'''
                            list_of_fittings = np.append(
                                list_of_fittings, np.array([[
                                    float(res.x[0]), res.x[1], abs(res.x[2]),
                                    self.gaussian_def.cal_chi2_multigaussian(res.x, spectrum_window2, xaxis_window2)
                                    / res.x[1] / len(spectrum_window2), len(spectrum_window2), "multiple",
                                    res.success]]), axis=0)

                            list_of_fittings = np.append(
                                list_of_fittings, np.array([[
                                    float(res.x[3]), res.x[4], abs(res.x[5]),
                                    self.gaussian_def.cal_chi2_multigaussian(res.x, spectrum_window2, xaxis_window2)
                                    / res.x[4] / len(spectrum_window2), len(spectrum_window2), "multiple",
                                    res.success]]), axis=0)

                            '''separated peak'''
                            res = self.gaussian_def.gaussian_fit_one_element_separated(
                                xaxis_window2, spectrum_window2,
                                local_max[0], spectrum_window2[np.where(local_max[0] == xaxis_window2)[0][0]])

                            if save_fitting:
                                self.plotting.plot_one_gaussian_separated(
                                    xaxis_window2, spectrum_window2,
                                    local_max[0], spectrum_window2[np.where(local_max[0] == xaxis_window2)[0][0]],
                                    res, path_outputs)

                            '''Add parameters to final table'''
                            list_of_fittings = np.append(
                                list_of_fittings, np.array([[
                                    float(res.x[0]), res.x[1], abs(res.x[2]),
                                    self.gaussian_def.cal_chi2(res.x, spectrum_window2, xaxis_window2)
                                    / res.x[1] / len(spectrum_window2), len(spectrum_window2), "separated",
                                    res.success]]), axis=0)

                        # The three peaks are independent and therefore fitted independently
                        if np.abs(local_max[1] - local_max[0]) >= 15 and np.abs(local_max[2] - local_max[1]) > 15:
                            for j in range(0, 3, 1):
                                peak_max = spectrum_window2[np.where(local_max[j] == xaxis_window2)]
                                xaxis_max = local_max[j]

                                res = self.gaussian_def.gaussian_fit_one_element_separated(
                                    xaxis_window2, spectrum_window2, xaxis_max, peak_max)

                                if save_fitting:
                                    self.plotting.plot_one_gaussian_separated(
                                        xaxis_window2, spectrum_window2, xaxis_max, peak_max, res, path_outputs)

                                '''Add parameters to final table'''
                                list_of_fittings = np.append(
                                    list_of_fittings, np.array([[
                                        float(res.x[0]), res.x[1], abs(res.x[2]),
                                        self.gaussian_def.cal_chi2(res.x, spectrum_window2, xaxis_window2)
                                        / res.x[1] / len(spectrum_window2), len(spectrum_window2), "separated",
                                        res.success]]), axis=0)
        return list_of_fittings

    def gaussian_fitting_spectrum_constrained(self, mean_spectrum, list_of_peaks, func, std, path_outputs,
                                              save_fitting=False):
        """
        Function for Gaussian fitting by chi2 minimization.
        The fitting constrains the Gaussian in centroid, intensity and width values.

        :param mean_spectrum: mean spectrum to take as reference to fit the Gaussian
        :type mean_spectrum: DataSpectrum
        :param list_of_peaks: list of peaks of interest (list created in peak_picking.py)
        :type list_of_peaks: list
        :param func: correlation function of the resolving power for the dataset
        :type func: poly1d
        :param std: intercept of the resolving power correlation for the dataset
        :type std: float
        :param path_outputs: outputs path
        :type path_outputs: string
        :param save_fitting: option to save plots of the Gaussian fitting (default = False)
        :type save_fitting: bool
        :return: table with the list of peaks and its characteristics
        (centroid, intensity, width, chi2/Nsamples, Number of sample bins, type of peak,
        success in the fitting, good peak shape)
        :rtype: array
        """

        if save_fitting:
            dirName = path_outputs + "Gaussian_fitting_width_constrained"
            if not os.path.exists(dirName):
                os.mkdir(dirName)

        mask = mean_spectrum.mzmask
        full_list_of_fittings = np.empty((0, 12))
        # 0              1          2               3                    4         5
        # left_min     meas mz   right_min   centroid gaussian    peak_intensity  width
        #     6              7            8       9      10                   11
        # chi2/Nsamples   Nsamples    Peak_type   S/N   Fitting_success    PeakShape"

        # It selects windows of signal to determine how many peaks are on them
        for i in range(2, len(mean_spectrum.intensities) - 2):
            if mask[i] == 0.0 and mask[i + 1] == 1.0:
                istart = i
            if mask[i] == 1.0 and mask[i + 1] == 0.0:
                spectrum_window = mean_spectrum.intensities[istart: i + 2]
                xaxis_window = mean_spectrum.mzs[istart: i + 2]

                local_maximum = []
                # It finds the number of local max in the window to determine how many
                # peaks are in the window.
                for ii in range(len(list_of_peaks)):
                    if max(xaxis_window) > list_of_peaks[ii] > min(xaxis_window):
                        local_maximum = np.append(local_maximum, list_of_peaks[ii])

                # If there is only one peak in the window it makes a single gaussian fitting
                '''Fits one peak'''
                if len(local_maximum) == 1:
                    xaxis_window_plot = []
                    spectrum_window_plot = []
                    centroid = np.argmin(np.abs(mean_spectrum.mzs - local_maximum))
                    for kkk in range((centroid - 50), (centroid + 50), 1):
                        xaxis_window_plot = np.append(xaxis_window_plot, mean_spectrum.mzs[kkk])
                        spectrum_window_plot = np.append(spectrum_window_plot, mean_spectrum.intensities[kkk])
                    res = self.gaussian_def.gaussian_fit_one_element_width_constraint(
                        xaxis_window_plot, spectrum_window_plot, local_maximum[0],
                        spectrum_window[np.where(local_maximum[0] == xaxis_window)[0][0]], func, std)
                    if save_fitting:
                        self.plotting.plot_one_gaussian_width_constraint(
                            xaxis_window_plot, spectrum_window_plot, local_maximum[0], res, path_outputs)

                    '''peak shape peak1'''
                    if (abs(res.x[2]) == func(local_maximum[0]) + 5 * std) or \
                            (abs(res.x[2]) == func(local_maximum[0]) - 5 * std):
                        flag = "bad"
                    else:
                        flag = "good"
                    '''Finds minima and local noise of peak1'''
                    left_min, right_min = self.dataop.find_minima_around_peak(
                        xaxis_window_plot, spectrum_window_plot, local_maximum[0])
                    # It estimates the local noise around the peak in a window of 20 Da
                    # to each size of the peak centroid
                    local_noise = self.noisedetermination.local_noise_around_peak(mean_spectrum, local_maximum[0],
                                                                                  window_size=20)
                    # It estimates the S/N as the peak intensity over the local noise
                    snr = res.x[1] / local_noise
                    '''Add parameters to final table'''
                    full_list_of_fittings = np.append(
                        full_list_of_fittings, np.array([[
                            left_min, local_maximum[0], right_min, float(res.x[0]), res.x[1], abs(res.x[2]),
                            self.gaussian_def.cal_chi2(res.x, spectrum_window, xaxis_window) / res.x[1]
                            / len(spectrum_window), len(spectrum_window), "single", snr, res.success, flag]]), axis=0)

                # If there are two peaks in the window, it determines if they are too
                # close to each other. It they are 15ppm apart, it fits each peak
                # as independent peaks. If not it fits them together as there might be
                # contamination between them.

                '''Fits two peaks'''
                if len(local_maximum) == 2:
                    if np.abs(local_maximum[1] - local_maximum[0]) >= 6 * func(local_maximum[0]):
                        '''Separated fit'''
                        for j in range(0, 2, 1):
                            peak_max = spectrum_window[np.where(local_maximum[j] == xaxis_window)[0][0]]
                            centroid = np.argmin(np.abs(mean_spectrum.mzs - local_maximum[j]))

                            xaxis_window_plot = []
                            spectrum_window_plot = []

                            for kkk in range((centroid - 50), (centroid + 50), 1):
                                xaxis_window_plot = np.append(xaxis_window_plot, mean_spectrum.mzs[kkk])
                                spectrum_window_plot = np.append(spectrum_window_plot, mean_spectrum.intensities[kkk])

                            res = self.gaussian_def.gaussian_fit_one_element_width_constraint(
                                xaxis_window_plot, spectrum_window_plot, local_maximum[j], peak_max, func, std)
                            if save_fitting:
                                self.plotting.plot_one_gaussian_separated_width_constraint(
                                    xaxis_window_plot, spectrum_window_plot, local_maximum[j], peak_max, res, func,
                                    std, path_outputs)

                            '''peak shape peak_j'''
                            if (abs(res.x[2]) == func(local_maximum[j]) + 5 * std) or \
                                    (abs(res.x[2]) == func(local_maximum[j]) - 5 * std):
                                flag = "bad"
                            else:
                                flag = "good"
                            '''Find minima and local noise around peak_j'''
                            left_min, right_min = self.dataop.find_minima_around_peak(
                                xaxis_window_plot, spectrum_window_plot, local_maximum[j])
                            local_noise = self.noisedetermination.local_noise_around_peak(
                                mean_spectrum, local_maximum[j], window_size=20)
                            snr = res.x[1] / local_noise
                            '''Add parameters to final table peak_j'''
                            full_list_of_fittings = np.append(
                                full_list_of_fittings, np.array([[
                                    left_min, local_maximum[j], right_min, float(res.x[0]), res.x[1], abs(res.x[2]),
                                    self.gaussian_def.cal_chi2(res.x, spectrum_window, xaxis_window) / res.x[1]
                                    / len(spectrum_window), len(spectrum_window), "single", snr, res.success, flag]]),
                                axis=0)

                    else:  # If they are close to each other, it fits them together.
                        '''Two fits together'''
                        xaxis_window_plot = []
                        spectrum_window_plot = []
                        centroid = np.argmin(np.abs(mean_spectrum.mzs - local_maximum[0]))

                        for kkk in range((centroid - 150), (centroid + 150), 1):
                            xaxis_window_plot = np.append(xaxis_window_plot, mean_spectrum.mzs[kkk])
                            spectrum_window_plot = np.append(spectrum_window_plot, mean_spectrum.intensities[kkk])

                        res = self.gaussian_def.gaussian_fit_two_elements_width_constraint(
                            xaxis_window_plot, spectrum_window_plot,
                            local_maximum[0], spectrum_window[np.where(local_maximum[0] == xaxis_window)[0][0]],
                            local_maximum[1], spectrum_window[np.where(local_maximum[1] == xaxis_window)[0][0]],
                            func, std)

                        if save_fitting:
                            self.plotting.plot_two_gaussians_width_constraint(
                                xaxis_window_plot, spectrum_window_plot,
                                local_maximum[0], spectrum_window[np.where(local_maximum[0] == xaxis_window)[0][0]],
                                local_maximum[1], spectrum_window[np.where(local_maximum[1] == xaxis_window)[0][0]],
                                func, std, res, path_outputs)

                        '''peak shape peak1'''
                        if (abs(res.x[2]) == func(local_maximum[0]) + 5 * std) or \
                                (abs(res.x[2]) == func(local_maximum[0]) - 5 * std):
                            flag = "bad"
                        else:
                            flag = "good"
                        '''Find minima and local noise around peak1'''
                        left_min, right_min = self.dataop.find_minima_around_peak(
                            xaxis_window_plot, spectrum_window_plot, local_maximum[0])
                        local_noise = self.noisedetermination.local_noise_around_peak(
                            mean_spectrum, local_maximum[0], window_size=20)
                        snr = res.x[1] / local_noise
                        '''Add parameters to final table peak1'''
                        full_list_of_fittings = np.append(
                            full_list_of_fittings, np.array([[
                                left_min, local_maximum[0], right_min, float(res.x[0]), res.x[1], abs(res.x[2]),
                                self.gaussian_def.cal_chi2_multigaussian(res.x, spectrum_window, xaxis_window)
                                / res.x[1] / len(spectrum_window), len(spectrum_window), "multiple", snr,
                                res.success, flag]]), axis=0)

                        '''peak shape peak2'''
                        if (abs(res.x[5]) == func(local_maximum[1]) + 5 * std) or \
                                (abs(res.x[5]) == func(local_maximum[1]) - 5 * std):
                            flag = "bad"
                        else:
                            flag = "good"
                        '''Find minima and local noise around peak2'''
                        left_min, right_min = self.dataop.find_minima_around_peak(
                            xaxis_window_plot, spectrum_window_plot, local_maximum[1])
                        local_noise = self.noisedetermination.local_noise_around_peak(
                            mean_spectrum, local_maximum[1], window_size=20)
                        snr = res.x[4] / local_noise
                        '''Add parameters to final table peak2'''
                        full_list_of_fittings = np.append(
                            full_list_of_fittings, np.array([[
                                left_min, local_maximum[1], right_min, float(res.x[3]), res.x[4], abs(res.x[5]),
                                self.gaussian_def.cal_chi2_multigaussian(res.x, spectrum_window, xaxis_window)
                                / res.x[4] / len(spectrum_window), len(spectrum_window), "multiple", snr,
                                res.success, flag]]), axis=0)

                # If there are tree peaks in the window, it determines the distance among
                # them and fit them accordingly.
                '''Fits three peaks'''
                if len(local_maximum) == 3:
                    # The three peaks are close to each other, they are fitted together
                    '''Three peaks together'''
                    if np.abs(local_maximum[1] - local_maximum[0]) <= 6 * func(local_maximum[0]) \
                            and np.abs(local_maximum[2] - local_maximum[1]) <= 6 * func(local_maximum[1]):

                        xaxis_window_plot = []
                        spectrum_window_plot = []
                        centroid = np.argmin(np.abs(mean_spectrum.mzs - local_maximum[1]))

                        for kkk in range((centroid - 150), (centroid + 150), 1):
                            xaxis_window_plot = np.append(xaxis_window_plot, mean_spectrum.mzs[kkk])
                            spectrum_window_plot = np.append(spectrum_window_plot, mean_spectrum.intensities[kkk])

                        res = self.gaussian_def.gaussian_fit_three_elements_width_constraint(
                            xaxis_window_plot, spectrum_window_plot,
                            local_maximum[0], spectrum_window[np.where(local_maximum[0] == xaxis_window)[0][0]],
                            local_maximum[1], spectrum_window[np.where(local_maximum[1] == xaxis_window)[0][0]],
                            local_maximum[2], spectrum_window[np.where(local_maximum[2] == xaxis_window)[0][0]],
                            func, std)

                        if save_fitting:
                            self.plotting.plot_three_gausssians_width_constraint(
                                xaxis_window_plot, spectrum_window_plot,
                                local_maximum[0], spectrum_window[np.where(local_maximum[0] == xaxis_window)[0][0]],
                                local_maximum[1], spectrum_window[np.where(local_maximum[1] == xaxis_window)[0][0]],
                                local_maximum[2], spectrum_window[np.where(local_maximum[2] == xaxis_window)[0][0]],
                                func, std, res, path_outputs)

                        '''peak shape peak1'''
                        if (abs(res.x[2]) == func(local_maximum[0]) + 5 * std) or \
                                (abs(res.x[2]) == func(local_maximum[0]) - 5 * std):
                            flag = "bad"
                        else:
                            flag = "good"
                        '''Find minima and local noise around peak1'''
                        left_min, right_min = self.dataop.find_minima_around_peak(
                            xaxis_window_plot, spectrum_window_plot, local_maximum[0])
                        local_noise = self.noisedetermination.local_noise_around_peak(
                            mean_spectrum, local_maximum[0], window_size=20)
                        snr = res.x[1] / local_noise
                        '''Add parameters to final table peak1'''
                        full_list_of_fittings = np.append(
                            full_list_of_fittings, np.array([[
                                left_min, local_maximum[0], right_min, float(res.x[0]), res.x[1], abs(res.x[2]),
                                self.gaussian_def.cal_chi2_multigaussian_3g(res.x, spectrum_window, xaxis_window)
                                / res.x[1] / len(spectrum_window), len(spectrum_window), "multiple",
                                snr, res.success, flag]]), axis=0)

                        '''peak shape peak2'''
                        if (abs(res.x[5]) == func(local_maximum[1]) + 5 * std) or \
                                (abs(res.x[5]) == func(local_maximum[1]) - 5 * std):
                            flag = "bad"
                        else:
                            flag = "good"
                        '''Find minima and local noise around peak2'''
                        left_min, right_min = self.dataop.find_minima_around_peak(
                            xaxis_window_plot, spectrum_window_plot, local_maximum[1])
                        local_noise = self.noisedetermination.local_noise_around_peak(
                            mean_spectrum, local_maximum[1], window_size=20)
                        snr = res.x[4] / local_noise
                        '''Add parameters to final table peak2'''
                        full_list_of_fittings = np.append(
                            full_list_of_fittings, np.array([[
                                left_min, local_maximum[1], right_min, float(res.x[3]), res.x[4], abs(res.x[5]),
                                self.gaussian_def.cal_chi2_multigaussian_3g(res.x, spectrum_window, xaxis_window)
                                / res.x[4] / len(spectrum_window), len(spectrum_window), "multiple",
                                snr, res.success, flag]]), axis=0)

                        '''peak shape peak3'''
                        if (abs(res.x[8]) == func(local_maximum[2]) + 5 * std) or \
                                (abs(res.x[8]) == func(local_maximum[2]) - 5 * std):
                            flag = "bad"
                        else:
                            flag = "good"
                        '''Find minima and local noise around peak3'''
                        left_min, right_min = self.dataop.find_minima_around_peak(
                            xaxis_window_plot, spectrum_window_plot, local_maximum[2])
                        local_noise = self.noisedetermination.local_noise_around_peak(
                            mean_spectrum, local_maximum[2], window_size=20)
                        snr = res.x[7] / local_noise
                        '''Add parameters to final table peak3'''
                        full_list_of_fittings = np.append(
                            full_list_of_fittings, np.array([[
                                left_min, local_maximum[2], right_min, float(res.x[6]), res.x[7], abs(res.x[8]),
                                self.gaussian_def.cal_chi2_multigaussian_3g(res.x, spectrum_window, xaxis_window)
                                / res.x[7] / len(spectrum_window), len(spectrum_window), "multiple",
                                snr, res.success, flag]]), axis=0)

                    '''Two peaks together + separated'''
                    # The first two peaks are together, they are fitted together. the last one is separated.
                    if np.abs(local_maximum[1] - local_maximum[0]) <= 6 * func(local_maximum[0]) \
                            and np.abs(local_maximum[2] - local_maximum[1]) > 6 * func(local_maximum[1]):

                        xaxis_window_plot_2 = []
                        spectrum_window_plot_2 = []
                        centroid = np.argmin(np.abs(mean_spectrum.mzs - local_maximum[1]))

                        for kkk in range((centroid - 150), (centroid + 150), 1):
                            xaxis_window_plot_2 = np.append(xaxis_window_plot_2, mean_spectrum.mzs[kkk])
                            spectrum_window_plot_2 = np.append(spectrum_window_plot_2, mean_spectrum.intensities[kkk])

                        res = self.gaussian_def.gaussian_fit_two_elements_width_constraint(
                            xaxis_window_plot_2, spectrum_window_plot_2,
                            local_maximum[0], spectrum_window[np.where(local_maximum[0] == xaxis_window)[0][0]],
                            local_maximum[1], spectrum_window[np.where(local_maximum[1] == xaxis_window)[0][0]],
                            func, std)

                        if save_fitting:
                            self.plotting.plot_two_gaussians_width_constraint(
                                xaxis_window_plot_2, spectrum_window_plot_2,
                                local_maximum[0], spectrum_window[np.where(local_maximum[0] == xaxis_window)[0][0]],
                                local_maximum[1], spectrum_window[np.where(local_maximum[1] == xaxis_window)[0][0]],
                                func, std, res, path_outputs)

                        '''peak shape peak1'''
                        if (abs(res.x[2]) == func(local_maximum[0]) + 5 * std) or \
                                (abs(res.x[2]) == func(local_maximum[0]) - 5 * std):
                            flag = "bad"
                        else:
                            flag = "good"
                        '''Find minima and local noise around peak1'''
                        left_min, right_min = self.dataop.find_minima_around_peak(
                            xaxis_window_plot_2, spectrum_window_plot_2, local_maximum[0])
                        local_noise = self.noisedetermination.local_noise_around_peak(
                            mean_spectrum, local_maximum[0], window_size=20)
                        snr = res.x[1] / local_noise
                        '''Add parameters to final table peak1'''
                        full_list_of_fittings = np.append(
                            full_list_of_fittings, np.array([[
                                left_min, local_maximum[0], right_min, float(res.x[0]), res.x[1], abs(res.x[2]),
                                self.gaussian_def.cal_chi2_multigaussian(res.x, spectrum_window, xaxis_window)
                                / res.x[1] / len(spectrum_window), len(spectrum_window), "multiple",
                                snr, res.success, flag]]), axis=0)

                        '''peak shape peak2'''
                        if (abs(res.x[5]) == func(local_maximum[1]) + 5 * std) or \
                                (abs(res.x[5]) == func(local_maximum[1]) - 5 * std):
                            flag = "bad"
                        else:
                            flag = "good"
                        '''Find minima and local noise around peak2'''
                        left_min, right_min = self.dataop.find_minima_around_peak(
                            xaxis_window_plot_2, spectrum_window_plot_2, local_maximum[1])
                        local_noise = self.noisedetermination.local_noise_around_peak(
                            mean_spectrum, local_maximum[1], window_size=20)
                        snr = res.x[4] / local_noise
                        '''Add parameters to final table peak2'''
                        full_list_of_fittings = np.append(
                            full_list_of_fittings, np.array([[
                                left_min, local_maximum[1], right_min, float(res.x[3]), res.x[4], abs(res.x[5]),
                                self.gaussian_def.cal_chi2_multigaussian(res.x, spectrum_window, xaxis_window)
                                / res.x[4] / len(spectrum_window), len(spectrum_window), "multiple",
                                snr, res.success, flag]]), axis=0)

                        '''Separated fitting'''
                        xaxis_window_plot = []
                        spectrum_window_plot = []
                        centroid = np.argmin(np.abs(mean_spectrum.mzs - local_maximum[2]))

                        for kkk in range((centroid - 50), (centroid + 50), 1):
                            xaxis_window_plot = np.append(xaxis_window_plot, mean_spectrum.mzs[kkk])
                            spectrum_window_plot = np.append(spectrum_window_plot, mean_spectrum.intensities[kkk])

                        res = (self.gaussian_def.gaussian_fit_one_element_width_constraint(
                            xaxis_window_plot, spectrum_window_plot,
                            local_maximum[2], spectrum_window[np.where(local_maximum[2] == xaxis_window)[0][0]],
                            func, std))

                        if save_fitting:
                            self.plotting.plot_one_gaussian_separated_width_constraint(
                                xaxis_window_plot, spectrum_window_plot,
                                local_maximum[2], spectrum_window[np.where(local_maximum[2] == xaxis_window)[0][0]],
                                res, func, std, path_outputs)

                        '''peak shape peak3'''
                        if (abs(res.x[2]) == func(local_maximum[2]) + 5 * std) or \
                                (abs(res.x[2]) == func(local_maximum[2]) - 5 * std):
                            flag = "bad"
                        else:
                            flag = "good"
                        '''Find minima and local noise around peak3'''
                        left_min, right_min = self.dataop.find_minima_around_peak(
                            xaxis_window_plot_2, spectrum_window_plot_2, local_maximum[2])
                        local_noise = self.noisedetermination.local_noise_around_peak(
                            mean_spectrum, local_maximum[2], window_size=20)
                        snr = res.x[1] / local_noise
                        '''Add parameters to final table peak3'''
                        full_list_of_fittings = np.append(
                            full_list_of_fittings, np.array([[
                                left_min, local_maximum[2], right_min, float(res.x[0]), res.x[1], abs(res.x[2]),
                                self.gaussian_def.cal_chi2(res.x, spectrum_window, xaxis_window)
                                / res.x[1] / len(spectrum_window), len(spectrum_window), "separated",
                                snr, res.success, flag]]), axis=0)

                    '''Separated + two peaks together'''
                    # The last two peaks are together, they are fitted together. the first one is separated.
                    if np.abs(local_maximum[2] - local_maximum[1]) <= 6 * func(local_maximum[1])\
                            and np.abs(local_maximum[1] - local_maximum[0]) > 6 * func(local_maximum[0]):
                        '''Two gaussian fit'''
                        xaxis_window_plot_2 = []
                        spectrum_window_plot_2 = []
                        centroid = np.argmin(np.abs(mean_spectrum.mzs - local_maximum[1]))

                        for kkk in range((centroid - 50), (centroid + 50), 1):
                            xaxis_window_plot_2 = np.append(xaxis_window_plot_2, mean_spectrum.mzs[kkk])
                            spectrum_window_plot_2 = np.append(spectrum_window_plot_2, mean_spectrum.intensities[kkk])

                        res = self.gaussian_def.gaussian_fit_two_elements_width_constraint(
                            xaxis_window_plot_2, spectrum_window_plot_2,
                            local_maximum[1], spectrum_window[np.where(local_maximum[1] == xaxis_window)[0][0]],
                            local_maximum[2], spectrum_window[np.where(local_maximum[2] == xaxis_window)[0][0]],
                            func, std)

                        if save_fitting:
                            self.plotting.plot_two_gaussians_width_constraint(
                                xaxis_window_plot_2, spectrum_window_plot_2,
                                local_maximum[1], spectrum_window[np.where(local_maximum[1] == xaxis_window)[0][0]],
                                local_maximum[2], spectrum_window[np.where(local_maximum[2] == xaxis_window)[0][0]],
                                func, std, res, path_outputs)

                        '''peak shape peak2'''
                        if (abs(res.x[2]) == func(local_maximum[1]) + 3 * std) \
                                or (abs(res.x[2]) == func(local_maximum[1]) - 3 * std):
                            flag = "bad"
                        else:
                            flag = "good"
                        '''Find minima and local noise around peak2'''
                        left_min, right_min = self.dataop.find_minima_around_peak(
                            xaxis_window_plot_2, spectrum_window_plot_2, local_maximum[1])
                        local_noise = self.noisedetermination.local_noise_around_peak(
                            mean_spectrum, local_maximum[1], window_size=20)
                        snr = res.x[1] / local_noise
                        '''Add parameters to final table peak2'''
                        full_list_of_fittings = np.append(
                            full_list_of_fittings, np.array([[
                                left_min, local_maximum[1], right_min, float(res.x[0]), res.x[1], abs(res.x[2]),
                                self.gaussian_def.cal_chi2_multigaussian(res.x, spectrum_window, xaxis_window)
                                / res.x[1] / len(spectrum_window), len(spectrum_window), "multiple",
                                snr, res.success, flag]]), axis=0)

                        '''peak shape peak3'''
                        if (abs(res.x[5]) == func(local_maximum[2]) + 5 * std) or \
                                (abs(res.x[5]) == func(local_maximum[2]) - 5 * std):
                            flag = "bad"
                        else:
                            flag = "good"
                        '''Find minima and local noise around peak3'''
                        left_min, right_min = self.dataop.find_minima_around_peak(
                            xaxis_window_plot_2, spectrum_window_plot_2, local_maximum[2])
                        local_noise = self.noisedetermination.local_noise_around_peak(
                            mean_spectrum, local_maximum[2], window_size=20)
                        snr = res.x[4] / local_noise
                        '''Add parameters to final table peak3'''
                        full_list_of_fittings = np.append(
                            full_list_of_fittings, np.array([[
                                left_min, local_maximum[2], right_min, float(res.x[3]), res.x[4], abs(res.x[5]),
                                self.gaussian_def.cal_chi2_multigaussian(res.x, spectrum_window, xaxis_window)
                                / res.x[4] / len(spectrum_window), len(spectrum_window), "multiple",
                                snr, res.success, flag]]), axis=0)

                        '''Separated peak - peak1'''
                        xaxis_window_plot = []
                        spectrum_window_plot = []
                        centroid = np.argmin(np.abs(mean_spectrum.mzs - local_maximum[0]))
                        for kkk in range((centroid - 50), (centroid + 50), 1):
                            xaxis_window_plot = np.append(xaxis_window_plot, mean_spectrum.mzs[kkk])
                            spectrum_window_plot = np.append(spectrum_window_plot, mean_spectrum.intensities[kkk])

                        res = (self.gaussian_def.gaussian_fit_one_element_width_constraint(
                            xaxis_window_plot_2, spectrum_window_plot_2,
                            local_maximum[0], spectrum_window[np.where(local_maximum[0] == xaxis_window)[0][0]],
                            func, std))

                        if save_fitting:
                            self.plotting.plot_one_gaussian_separated_width_constraint(
                                xaxis_window_plot_2, spectrum_window_plot_2,
                                local_maximum[0], spectrum_window[np.where(local_maximum[0] == xaxis_window)[0][0]],
                                res, func, std, path_outputs)

                        '''peak shape peak1'''
                        if (abs(res.x[2]) == func(local_maximum[0]) + 5 * std) \
                                or (abs(res.x[2]) == func(local_maximum[0]) - 5 * std):
                            flag = "bad"
                        else:
                            flag = "good"
                        '''Find minima and local noise around peak1'''
                        left_min, right_min = self.dataop.find_minima_around_peak(
                            xaxis_window_plot_2, spectrum_window_plot_2, local_maximum[0])
                        local_noise = self.noisedetermination.local_noise_around_peak(
                            mean_spectrum, local_maximum[0], window_size=20)
                        snr = res.x[1] / local_noise
                        '''Add parameters to final table peak1'''
                        full_list_of_fittings = np.append(
                            full_list_of_fittings, np.array([[
                                left_min, local_maximum[0], right_min, float(res.x[0]), res.x[1], abs(res.x[2]),
                                self.gaussian_def.cal_chi2(res.x, spectrum_window, xaxis_window) / res.x[1]
                                / len(spectrum_window), len(spectrum_window), "separated", snr, res.success, flag]]),
                            axis=0)

                '''More than 3 peaks on window'''
                if len(local_maximum) > 3:
                    local_maximum = np.append(local_maximum, 0)
                    for j in range(len(local_maximum) - 1):
                        if j == 0:
                            xaxis_window_plot = []
                            spectrum_window_plot = []
                            centroid = np.argmin(np.abs(mean_spectrum.mzs - local_maximum[j]))

                            for kkk in range((centroid - 80), (centroid + 80), 1):
                                xaxis_window_plot = np.append(xaxis_window_plot, mean_spectrum.mzs[kkk])
                                spectrum_window_plot = np.append(spectrum_window_plot, mean_spectrum.intensities[kkk])

                            res = self.gaussian_def.gaussian_fit_two_elements_width_constraint(
                                xaxis_window_plot, spectrum_window_plot,
                                local_maximum[j], spectrum_window[np.where(local_maximum[j] == xaxis_window)[0][0]],
                                local_maximum[j + 1], spectrum_window[np.where(local_maximum[j + 1] ==
                                                                               xaxis_window)[0][0]],
                                func, std)

                            if save_fitting:
                                self.plotting.plot_two_gaussians_width_constraint(
                                    xaxis_window_plot, spectrum_window_plot,
                                    local_maximum[j], spectrum_window[np.where(local_maximum[j] == xaxis_window)[0][0]],
                                    local_maximum[j + 1], spectrum_window[np.where(local_maximum[j + 1] ==
                                                                                   xaxis_window)[0][0]],
                                    func, std, res, path_outputs)

                            '''peak shape peak_j'''
                            if (abs(res.x[2]) == func(local_maximum[j]) + 5 * std) \
                                    or (abs(res.x[2]) == func(local_maximum[j]) - 5 * std):
                                flag = "bad"
                            else:
                                flag = "good"
                            '''Find minima and local noise around peak_j'''
                            left_min, right_min = self.dataop.find_minima_around_peak(
                                xaxis_window_plot, spectrum_window_plot, local_maximum[j])
                            local_noise = self.noisedetermination.local_noise_around_peak(
                                mean_spectrum, local_maximum[j], window_size=20)
                            snr = res.x[1] / local_noise
                            '''Add parameters to final table peak_j'''
                            full_list_of_fittings = np.append(
                                full_list_of_fittings, np.array([[
                                    left_min, local_maximum[j], right_min, float(res.x[0]), res.x[1], abs(res.x[2]),
                                    self.gaussian_def.cal_chi2_multigaussian(res.x, spectrum_window, xaxis_window)
                                    / res.x[1] / len(spectrum_window), len(spectrum_window), "multiple",
                                    snr, res.success, flag]]), axis=0)

                        elif j == (len(local_maximum) - 2):
                            xaxis_window_plot = []
                            spectrum_window_plot = []
                            centroid = np.argmin(np.abs(mean_spectrum.mzs - local_maximum[j]))

                            for kkk in range((centroid - 80), (centroid + 80), 1):
                                xaxis_window_plot = np.append(xaxis_window_plot, mean_spectrum.mzs[kkk])
                                spectrum_window_plot = np.append(spectrum_window_plot, mean_spectrum.intensities[kkk])

                            res = self.gaussian_def.gaussian_fit_two_elements_width_constraint(
                                xaxis_window_plot, spectrum_window_plot,
                                local_maximum[j - 1], spectrum_window[np.where(local_maximum[j - 1] ==
                                                                               xaxis_window)[0][0]],
                                local_maximum[j], spectrum_window[np.where(local_maximum[j] == xaxis_window)[0][0]],
                                func, std)

                            if save_fitting:
                                self.plotting.plot_two_gaussians_width_constraint(
                                    xaxis_window_plot, spectrum_window_plot,
                                    local_maximum[j - 1], spectrum_window[np.where(local_maximum[j - 1] ==
                                                                                   xaxis_window)[0][0]],
                                    local_maximum[j], spectrum_window[np.where(local_maximum[j] == xaxis_window)[0][0]],
                                    func, std, res, path_outputs)

                            '''peak shape peak_j'''
                            if (abs(res.x[5]) == func(local_maximum[j]) + 5 * std) \
                                    or (abs(res.x[5]) == func(local_maximum[j]) - 5 * std):
                                flag = "bad"
                            else:
                                flag = "good"
                            '''Add parameters to final table peak_j'''
                            left_min, right_min = self.dataop.find_minima_around_peak(
                                xaxis_window_plot, spectrum_window_plot, local_maximum[j])
                            local_noise = self.noisedetermination.local_noise_around_peak(
                                mean_spectrum, local_maximum[j], window_size=30)
                            snr = res.x[4] / local_noise
                            '''Add parameters to final table peak_j'''
                            full_list_of_fittings = np.append(
                                full_list_of_fittings, np.array([[
                                    left_min, local_maximum[j], right_min, float(res.x[3]), res.x[4], abs(res.x[5]),
                                    self.gaussian_def.cal_chi2_multigaussian(res.x, spectrum_window, xaxis_window)
                                    / res.x[4] / len(spectrum_window), len(spectrum_window), "multiple",
                                    snr, res.success, flag]]), axis=0)

                        else:
                            xaxis_window_plot = []
                            spectrum_window_plot = []
                            centroid = np.argmin(np.abs(mean_spectrum.mzs - local_maximum[j]))

                            for kkk in range((centroid - 80), (centroid + 80), 1):
                                xaxis_window_plot = np.append(xaxis_window_plot, mean_spectrum.mzs[kkk])
                                spectrum_window_plot = np.append(spectrum_window_plot, mean_spectrum.intensities[kkk])

                            res = self.gaussian_def.gaussian_fit_three_elements_width_constraint(
                                xaxis_window_plot, spectrum_window_plot,
                                local_maximum[j - 1], spectrum_window[np.where(local_maximum[j - 1] ==
                                                                               xaxis_window)[0][0]],
                                local_maximum[j], spectrum_window[np.where(local_maximum[j] ==
                                                                           xaxis_window)[0][0]],
                                local_maximum[j + 1], spectrum_window[np.where(local_maximum[j + 1] ==
                                                                               xaxis_window)[0][0]],
                                func, std)

                            if save_fitting:
                                self.plotting.plot_three_gausssians_width_constraint(
                                    xaxis_window_plot, spectrum_window_plot,
                                    local_maximum[j - 1], spectrum_window[np.where(local_maximum[j - 1] ==
                                                                                   xaxis_window)[0][0]],
                                    local_maximum[j], spectrum_window[np.where(local_maximum[j] ==
                                                                               xaxis_window)[0][0]],
                                    local_maximum[j + 1], spectrum_window[np.where(local_maximum[j + 1] ==
                                                                                   xaxis_window)[0][0]],
                                    func, std, res, path_outputs)

                            '''peak shape peak_j'''
                            if (abs(res.x[5]) == func(local_maximum[j]) + 5 * std) \
                                    or (abs(res.x[5]) == func(local_maximum[j]) - 5 * std):
                                flag = "bad"
                            else:
                                flag = "good"
                            '''Add parameters to final table peak_j'''
                            left_min, right_min = self.dataop.find_minima_around_peak(
                                xaxis_window_plot, spectrum_window_plot, local_maximum[j])
                            local_noise = self.noisedetermination.local_noise_around_peak(
                                mean_spectrum, local_maximum[j], window_size=30)
                            snr = res.x[4] / local_noise
                            '''Add parameters to final table peak_j'''
                            full_list_of_fittings = np.append(
                                full_list_of_fittings, np.array([[
                                    left_min, local_maximum[j], right_min, float(res.x[3]), res.x[4], abs(res.x[5]),
                                    self.gaussian_def.cal_chi2_multigaussian_3g(res.x, spectrum_window, xaxis_window)
                                    / res.x[4] / len(spectrum_window), len(spectrum_window), "multiple",
                                    snr, res.success, flag]]), axis=0)

        return full_list_of_fittings
